import re
import warnings

FT_DEPTH = 999

class InfiniteLoopError(Exception):
    pass

class FTGroupsCompileError(Exception):
    pass


class FTMatched:
    def __init__(self, matched:str, group0:str, string:str, pattern:str, span:tuple, groups:tuple, areas:tuple, *, expansion=None):

        self._string = string
        self._group0 = group0
        self._matched = matched
        self._pattern = pattern
        self._span = tuple(span)
        self._groups = tuple(groups)
        self._areas = tuple(areas)

        assert matched is not None, f'Matched can not be None.'

        # calc once
        self._lineno = self._string[:self._span[0]].count('\n') + 1
        self._lineno = self._calc_expansion(self._lineno, expansion)
        self._hash = hash((self._string, self._group0, self._matched, self._pattern, self._span, self._groups, self._areas))

    @staticmethod
    def _calc_expansion(no, expansion):
        """
        Calculate the expansion
        """
        if expansion is None:
            return no
        for k, v in expansion.items():
            if no > k:
                no -= v
        return no

    @property
    def string(self):
        return self._string

    @property
    def group0(self):
        return self._group0

    @property
    def matched(self):
        return self._matched

    @property
    def pattern(self):
        return self._pattern

    @property
    def span(self):
        return self._span

    @property
    def groups(self):
        return self._groups

    @property
    def areas(self):
        return self._areas

    # ex
    @property
    def lineno(self):
        """
        Return the line number of the matched. Start from 1
        """
        return self._lineno

    @property
    def lineidx(self):
        """
        Return the index of the matched in the line. Start from 0
        """
        return self._lineno - 1


    def __str__(self):
        return self._matched

    def __int__(self):
        return int(self._matched)

    def __float__(self):
        return float(self._matched)

    def __bool__(self):
        return bool(self._matched)

    def __repr__(self):
        return self._matched

    def __getitem__(self, item):
        return self._matched[item]

    def __len__(self):
        return len(self._matched)

    def __iter__(self):
        return self._matched.__iter__()

    def __contains__(self, item):
        return item in self._matched

    def __eq__(self, other):
        return self._matched == other

    def __ne__(self, other):
        return self._matched != other

    def __add__(self, other):
        return self._matched + other

    def __radd__(self, other):
        return other + self._matched

    def __mul__(self, other):
        return self._matched * other

    def __rmul__(self, other):
        return other * self._matched

    def __hash__(self):
        return self._hash


class Current:
    def __init__(self, parent):
        self.parent = parent
        self.id = None
        self.func = None
        self.groups = None
        self.areas = None
        self.embed = None
        self.replace = None

        # ---
        self.expansions = {}   # {int:int}
        self._last_pattern = None
        self._last_pattern_id = None

        # --- not safe
        self.last_position = None
        self.raw_txt = None
        self.matched = None

    def clear(self):
        self.expansions = {}   # {int:int}
        self._last_pattern = None
        self._last_pattern_id = None
        self.last_position = None
        self.raw_txt = None
        self.matched = None

    @property
    def offset(self):
        return self.last_position if self.last_position is not None else 0

    @property
    def pattern(self):
        if self._last_pattern_id == self.id:
            return self._last_pattern
        self._last_pattern = ""
        self._last_pattern_id = self.id
        for group in self.groups:
            self._last_pattern += f'({group})'

        return self._last_pattern

    def toid(self, id):
        if id not in self.parent.ids:
            warnings.warn(f'to: Id {self.id} not exists')
            return

        index = self.parent.ids.index(self.id)
        self.id = id
        self.func = self.parent.funcs[index]
        self.groups = self.parent.groupss[index]
        self.areas = self.parent.areass[index]
        self.embed = self.parent.embeds[index]
        self.replace = self.parent.replaces[index]

        self.last_position = None
        self.matched = None

    def toindex(self, index):
        if index < 0 or index >= len(self.parent.ids):
            warnings.warn(f'to: Index {index} out of range')
            return

        self.id = self.parent.ids[index]
        self.func = self.parent.funcs[index]
        self.groups = self.parent.groupss[index]
        self.areas = self.parent.areass[index]
        self.embed = self.parent.embeds[index]
        self.replace = self.parent.replaces[index]

        self.last_position = None
        self.matched = None

    def __str__(self):
        return f'{self.__class__.__name__}: id={self.id}, func={self.func}, groups={self.groups}, areas={self.areas}, embed={self.embed}, replace={self.replace}'


class FTFlag:
    NONE = 0
    REDO = 1  # redo the current after current finished.  This wont reset the 'lefts'
    SKIP = 2  # skip the current and go to next one
    STOP = 4  # stop the handle process

# Fomater Tool
class FT:
    VARIABLE = "[a-zA-Z_][a-zA-Z0-9_]*"
    NUMBER = "[0-9]*\.?[0-9]+"
    ALPHA = "[a-zA-Z_][a-zA-Z]*"
    INTEGER = "[0-9]+"
    COMMA = ","
    COLON = ":"
    SEMICOLON = ";"
    EQUAL = "="
    PLUS = "\+"
    MINUS = "-"
    MULTIPLY = "\*"
    DIVIDE = "/"
    POWER = "\^"
    LBRACKET = "\("
    RBRACKET = "\)"
    LBRACE = "\{"
    RBRACE = "\}"
    LBRACKET_S = "\["
    RBRACKET_S = "\]"
    LANGLE = "<"
    RANGLE = ">"
    DOLLAR = "\$"
    DOT = "\."
    QUESTION = "\?"

    # RE Special
    BLANK_CHAR = "\s"
    DIGIT_CHAR = "\d"
    WORD_CHAR = "\w"
    NBLANK_CHAR = "\S"
    NDIGIT_CHAR = "\D"
    NWORD_CHAR = "\W"
    ANY_CHAR = "(.|\n)"
    BLANK_NOT_NEWLINE = "[ \t\r\f\v]"
    BNN = BLANK_NOT_NEWLINE
    OP_ANYMORE = "*"
    OP_ATLEAST = "+"
    OP_MAYBE = "?"

    # #
    BLANK_ANYMORE = BLANK_CHAR + OP_ANYMORE
    BLANK_ATLEAST = BLANK_CHAR + OP_ATLEAST
    BLANK_MAYBE = BLANK_CHAR + OP_MAYBE

    def __init__(self):
        self._id = -1  # id
        self.ids = []  # list of ids
        self.funcs = []  # list of functions
        self.groupss = []  # list of groups
        self.areass = []  # list of areas
        self.embeds = []  # list of (is_)embeds
        self.replaces = []  # list of (is_)replaces

        # current
        self.current = Current(self)
        self.flag = FTFlag.NONE

    def link(self, other) -> 'FT':
        """
        Link other FT to this FT
        """
        new_ft = FT()
        new_ft.ids = self.ids + other.ids
        new_ft.funcs = self.funcs + other.funcs
        new_ft.groupss = self.groupss + other.groupss
        new_ft.areass = self.areass + other.areass
        new_ft.embeds = self.embeds + other.embeds
        new_ft.replaces = self.replaces + other.replaces
        return new_ft

    @property
    def id(self):
        self._id += 1
        return self._id

    def _get_id(self):
        return id(self) * 10 + self.id

    @staticmethod
    def WrapForward(eq:bool, pattern:str) -> str:
        return f'?{"=" if eq else "!"}({pattern})'

    WF = WrapForward

    @staticmethod
    def WrapBackward(eq:bool, pattern:str) -> str:
        return f'?<{"=" if eq else "!"}({pattern})'

    WB = WrapBackward

    @staticmethod
    def ToParttern(*groups, clean=True) -> str:
        # clean groups
        if clean:
            groups = FT._api_clean_groups(groups)
        return ''.join([f'({group})' for group in groups])


    @staticmethod
    def ToPartternWithAreas(*groups, areas=None) -> tuple[str, list[int]]:
        # clean groups
        groups = FT._api_clean_groups(groups)

        # areas
        areas = FT._api_auto_prexpand_area(areas, groups)

        # rshift areas
        areas = FT._api_groups_bracket_rshift(groups, areas)

        return FT.ToParttern(*groups, clean=False), areas

    @staticmethod
    def Extract(txt: str, *gs, areas=None) -> list[tuple[FTMatched]]:
        """
        Extract all matched without replace
        """
        _res = []
        ft = FT()
        ft.login(lambda *args: _res.append(args), *gs, areas=areas, replace=False)
        ft.handle(str(txt))
        return _res

    @staticmethod
    def _check_bracket(txt: str):
        """
        Check if any brackets in txt
        :param txt:
        :return:
        """
        if txt.find('(') != -1 and txt.find(')') != -1:
            warnings.warn(f'Should not use brackets in {txt}')
            return True
        return False

    @staticmethod
    def _expand_area(areas, size) -> list:
        """
        Expand area to a list
        :param areas:
        :param size: max value. If index over size, will raise Exception
        :return:
        """
        res = []
        for area in areas:
            if isinstance(area, int):
                res += [area]
            elif isinstance(area, tuple):
                if len(area) != 2:
                    raise Exception(f'Area {area} should be a tuple of (start, end)')
                start, end = area
                if start < 0 or start >= size:
                    raise Exception(f'Area {area} start {start} out of range')
                if end < 0 or end >= size:
                    raise Exception(f'Area {area} end {end} out of range')
                if start > end:
                    raise Exception(f'Area {area} start {start} should less than end {end}')
                res += list(range(start, end))
            else:
                raise Exception(f'Area {area} should be a tuple of (start, end) or a index')
        return res


    @staticmethod
    def _api_rm_useless_bracket(txt: str) -> str:
        """
        Remove completely enclosed brackets at the beginning and end of the string, and these brackets should be matching pairs.
        :param txt: The input string with brackets.
        :return: The string with useless brackets removed.
        """

        def _inner_fn(_txt):
            stack, lens = [], len(_txt)
            try:
                for i, char in enumerate(_txt):
                    if char == '(' and (i == 0 or _txt[i - 1] != '\\'):
                        stack.append((char, i))
                    elif char == ')' and (i == 0 or _txt[i - 1] != '\\'):
                        if len(stack) == 1 and stack[0][1] == 0 and i == lens - 1:
                            # 如果堆栈中只有一个元素，且这个元素是字符串的第一个字符，且当前字符是字符串的最后一个字符
                            return _txt[1:-1], True
                        elif stack[-1][0] == '(':
                            stack.pop()
                        else:
                            raise IndexError(f"ft:Brackets do not match at {i} of {_txt}")
            except IndexError as e:
                raise ValueError(f'ft:Brackets do not match in {_txt}') from e
            return _txt, False

        while True:
            txt, changed = _inner_fn(txt)
            if not changed:
                break
        return txt

    @staticmethod
    def _api_count_bracket(txt: str) -> int:
        """
        Count the number of brackets without '\'
        * Ignore ?=, ?!, ?<=, ?<! and (
        :param txt:
        :return:
        """
        count = 0
        for i in range(len(txt)):
            if txt[i] == '(' and (i == 0 or txt[i - 1] != '\\'):
                # 排除正向前瞻和负向前瞻
                if i >= 3 and txt[i-3:i] in ['?<!', '?<=']:
                    continue
                if i >= 2 and txt[i-2:i] in ['?!', '?=']:
                    continue
                count += 1
        return count

    @staticmethod
    def _api_groups_bracket_rshift(gs, areas) -> list:
        """
        rshift the areas if a groups' element have brackets
        """
        lena, copya = len(areas), areas.copy()
        for i, g in enumerate(gs):
            br_cnt = FT._api_count_bracket(g)
            if br_cnt > 0:
                # while t < lena:
                for t in range(lena):
                    if areas[t] > i:  # 考虑areas中超过该index的情况
                        copya[t] += br_cnt
                        t += 1

        return copya


    @staticmethod
    def _api_auto_prexpand_area(areas, groups) -> list:
        # areas
        if areas is None:
            areas = list(range(len(groups)))
        else:
            areas = FT._expand_area(areas, len(groups))
        return areas

    @staticmethod
    def _api_clean_groups(groups) -> tuple:
        # clean groups
        _tmp_groups = list(groups)
        for i, group in enumerate(groups):
            try:
                _tmp_groups[i] = FT._api_rm_useless_bracket(group)
            except ValueError as e:
                raise ValueError(f'ft.login: Unmatched brackets in {group}') from e
        groups = tuple(_tmp_groups)
        return groups

    @staticmethod
    def _try_compile_groups(groups):
        pattern = FT.ToParttern(*groups)
        re.compile(pattern)     # if occur any error, will raise exception

    def login(self, fn, *groups: str, areas: list = None, embed:bool=False, replace:bool=True) -> int:
        """
        Login a function
        :param fn:
        :param groups: can be empty
        :param areas: list of area, each area is a tuple of (start, end) or a index
        :param embed: if embed, groups must have more than 3 items. And rfind first-element when matched.
        :param replace: if replace, will replace the old one
        :return: id

        * Special，you can only pass in a param:func. This will be regard as a simply pass function, and call it without any params.
        """
        # for group in groups:
        #     self._check_bracket(group)

        # check exists
        if fn in self.funcs:
            warnings.warn(f'Function {fn} already exists')

        # check groups
        if groups:  # not pass-in mode.
            if groups in self.groupss:
                warnings.warn(f'Groups {groups} already exists')

            # clean groups
            groups = self._api_clean_groups(groups)

            # areas
            areas = self._api_auto_prexpand_area(areas, groups)

            # rshift areas
            areas = self._api_groups_bracket_rshift(groups, areas)

            # embed
            if embed:
                if len(groups) < 3:
                    raise Exception(f'Groups at embed mode should have more than 3 elements.')
                if not replace:
                    raise Exception(f'Embed mode should use replace=True to avoid infinite loop.')

        # compile check
        try:
            self._try_compile_groups(groups)
        except Exception as e:
            raise FTGroupsCompileError(f'FT.login: Groups compile failed:\n\t{groups}\n\npattern:\n\t{self.ToParttern(*groups)}\n\ndetails:\n\t{e}') from e

        # add
        _id = self._get_id()
        self.ids.append(_id)
        self.funcs.append(fn)
        self.groupss.append(groups)
        self.areass.append(areas)
        self.embeds.append(embed)
        self.replaces.append(replace)

        return _id

    def logout(self, *ids: int):
        """
        logout fn or groups
        :param ids: id return from .login
        :return:
        """
        for identy in ids:
            if identy not in self.ids:
                warnings.warn(f'logout: Id {identy} not exists')
                continue
            index = self.ids.index(identy)
            self.ids.pop(index)
            self.funcs.pop(index)
            self.groupss.pop(index)
            self.areass.pop(index)
            self.embeds.pop(index)
            self.replaces.pop(index)


    @staticmethod
    def _tuple_fall_ele_to_str(fall_ele:tuple|str) -> str:
        """
        Convert fall element to str
        """
        if isinstance(fall_ele, tuple):
            # check only one element is not ''
            _cnt, _the = 0, ""
            for e in fall_ele:
                if e != '':
                    _cnt += 1
                    _the = e
            if _cnt != 1:
                raise Exception(f'Embed with tuple pats should have only one element not \'\'.\n\tget:\n\t\t{fall_ele}')
            return _the
        return fall_ele

    def _consistent_forward_match(self, pat: str, txt: str, full_check_pat:str) -> tuple:
        """
        一致性向前匹配目标，并要求匹配到的目标往后仍然符合整体匹配
        :param pat:
        :param txt:
        :param full_check_pat: 总体检查模式
        :return: full matched (start, end) or None
        """
        raw_txt = txt
        fall = re.findall(pat, txt)
        fall.reverse()
        cut_content = ""
        for index, ele in enumerate(fall):
            # 支持'|'分割导致的tuple格式的find
            ele = self._tuple_fall_ele_to_str(ele)
            _pos = txt.rfind(ele)
            if _pos == -1:
                raise Exception(f'Program meet an strange critical error. Re find a element:{ele}, but rfind report -1;')
            cut_content = txt[_pos:] + cut_content
            full_matched = re.fullmatch(full_check_pat, cut_content)
            if full_matched is not None:
                # _debug_str = raw_txt[_pos:len(raw_txt)]
                return _pos, len(raw_txt)
            txt = txt[:_pos]
        return None

    def _consistent_backward_match(self, pat: str, txt: str, full_check_pat:str) -> tuple:
        """
        一致性向后匹配目标，并要求匹配到的目标往前仍然符合整体匹配
        :param pat:
        :param txt:
        :param full_check_pat: 总体检查模式
        :return: full matched (start, end) or None
        """
        raw_txt = txt
        fall = re.findall(pat, txt)
        cut_content = ""
        for index, ele in enumerate(fall):
            # 支持'|'分割导致的tuple格式的find
            ele = self._tuple_fall_ele_to_str(ele)
            _pos = txt.find(ele)
            if _pos == -1:
                raise Exception(f'Program meet an strange critical error. Re find a element:{ele}, but rfind report -1;')
            cut_content += txt[:_pos + len(ele)]
            full_matched = re.fullmatch(full_check_pat, cut_content)
            if full_matched is not None:
                # _debug_str = raw_txt[:_pos + len(ele)]
                return 0, len(cut_content)
            txt = txt[_pos + len(ele):]
        return None


    def _handle_matched(self, matched:re.Match) -> str:
        # get each groups
        cares = matched.groups()
        # create params
        params = []
        for area in self.current.areas:
            if not isinstance(area, int):
                raise Exception(f'Area {self.current_areas} should be a list[int].')

            # area is index, index 0 is total, need +1 to get the correspond group
            _span = matched.span(area + 1)
            _span = (self.current.offset + _span[0], self.current.offset + _span[1])
            ftm = FTMatched(matched.group(area + 1), matched.group(), self.current.raw_txt, matched.re.pattern, _span, self.current.groups, self.current.areas, expansion=self.current.expansions)
            params.append(ftm)
        # call
        return str(self.current.func(*params))

    def _handle_current_once(self, raw_txt: str) -> str:
        """
        Handle current once.
        return changed_txt or None
        """

        self.current.raw_txt = raw_txt

        if self.current.last_position is not None:
            txt = raw_txt[self.current.last_position:]
            offset = self.current.last_position
        else:
            txt = raw_txt
            offset = 0

        # search
        matched = re.search(self.current.pattern, txt)


        # not matched
        if matched is None:
            self.current.last_position = None
            return

        span = (matched.start() + offset, matched.end() + offset)
        _raw_span = span

        # if embed, rfind first-element and find the neg-element
        if self.current.embed:
            gfe = self.current.groups[0]
            content = matched.group()
            cfm = self._consistent_forward_match(gfe, content, self.current.pattern)
            if cfm is not None:
                span = cfm

            gne = self.current.groups[-1]
            content = content[span[0]:]
            cbm = self._consistent_backward_match(gne, content, self.current.pattern)
            if cbm is not None:
                span = (span[0], cbm[1] + span[0])

            # update matched
            content = matched.group()[span[0]:span[1]]
            matched = re.fullmatch(self.current.pattern, content)

            # add bias refer to raw_txt
            span = (span[0] + _raw_span[0], span[1] + _raw_span[0])

        # general mode
        rep = self._handle_matched(matched)
        if self.current.replace:
            loop_check = re.search(self.current.pattern, rep)
            if loop_check is not None:
                warnings.warn(f'Infinite loop detected in your replace txt: {rep}. Which matched by {matched} in: \t{self.current}')

        # first exit point
        self.current.last_position = span[1] + 1
        if not self.current.replace:
            return raw_txt

        # replace
        txt = raw_txt[:span[0]] + rep + raw_txt[span[1]:]
        self.current.last_position = None  # reset if replace
        return txt


    def handle(self, txt: str, depth=FT_DEPTH):
        self.current.clear()    # clear current
        lens = len(self.funcs), len(self.groupss)
        assert lens[0] == lens[1], f'Length of funcs, groups not equal: {lens}'
        for i in range(lens[0]):
            self.current.toindex(i)
            lefts = depth

            while True:  # 完全由内部控制
                # check pass-in mode
                if len(self.current.groups) == 0:
                    # simply exec
                    self.current.func()
                else:
                    # loop-match
                    while lefts > 0:
                        lefts -= 1
                        ret = self._handle_current_once(txt)
                        if ret is None:
                            break
                        txt = ret

                        if self.flag & FTFlag.STOP:
                            # self.flag &= ~FTFlag.STOP  # Change at outer while
                            break

                        if self.flag & FTFlag.SKIP:
                            self.flag &= ~FTFlag.SKIP
                            break

                        if lefts == 0:
                            raise InfiniteLoopError(f'Infinite loop detected in {self.current}')


                # flag check
                if self.flag & FTFlag.STOP:
                    self.flag &= ~FTFlag.STOP
                    break
                elif self.flag & FTFlag.REDO:
                    self.flag &= ~FTFlag.REDO
                    continue
                else:
                    break

        return txt

    def logexpn(self, valid_no:int, bias_no:int):
        """
        记录 expansion信息
        * 只在当前handle周期内有效
        最终，当某个计算得到的no > valid_no时，将会-bias_no
        * 理论上, valid_no在同一周期内不会重复，具有唯一性
        """
        self.current.expansions[valid_no] = bias_no


    # call
    def __call__(self, txt: str):
        return self.handle(txt)

    def __str__(self):
        return f'{self.__class__.__name__}: funcs={self.funcs}, groups={self.groupss}, areas={self.areass}'

    def __len__(self):
        return len(self.funcs)

    def __iter__(self) -> Current:
        for i in range(len(self.funcs)):
            self.current.toindex(i)
            yield self.current

    def __add__(self, other):
        return self.link(other)

    def __iadd__(self, other):
        _new = self.link(other)
        self.ids = _new.ids
        self.funcs = _new.funcs
        self.groupss = _new.groupss
        self.areass = _new.areass
        self.embeds = _new.embeds
        self.replaces = _new.replaces
        return self

class FRemove(FT):
    """
    Remove data by format
    """

    def __init__(self):
        super().__init__()

    def login(self, *groups: str):
        super().login(lambda *args: "", *groups, replace=True)


class FDict(FT):
    """
    Get Dict data by format
    """

    def __init__(self):
        super().__init__()
        self.dict = {}

    def _hook_key_val(self, *args, key_id=None, val_id=None, val_default=None):
        assert key_id is not None, f'key_id should not be None'
        key = args[key_id]
        if val_id is None:
            val = val_default
        else:
            val = args[val_id]
        self.dict[key] = val
        return ''

    def login(self, key_id, val_id, *groups: str, areas: list = None, val_default=None, replace=False):
        fn = lambda *args: self._hook_key_val(*(str(it) for it in args), key_id=key_id, val_id=val_id, val_default=val_default)
        super().login(fn, *groups, areas=areas, replace=replace)

    # dict functions
    def keys(self):
        return self.dict.keys()

    def values(self):
        return self.dict.values()

    def items(self):
        return self.dict.items()

    def __getitem__(self, item):
        return self.dict[item]

    def __setitem__(self, key, value):
        self.dict[key] = value

    def __contains__(self, item):
        return item in self.dict

    def __len__(self):
        return len(self.dict)

    def __iter__(self):
        return self.dict.__iter__()

    def __str__(self):
        return self.dict.__str__()

    def __bool__(self):
        return bool(self.dict)

    def clear(self):
        self.dict.clear()

    def pop(self, key):
        return self.dict.pop(key)

    def get(self, key, default=None):
        return self.dict.get(key, default)

    def todict(self):
        return self.dict.copy()


class FSet(FT):
    """
    Get Set data by format
    """

    def __init__(self):
        super().__init__()
        self.set = set()

    def _hook_set(self, *args, val_id=None):
        assert val_id is not None, f'val_id should not be None'
        self.set.add(args[val_id])
        return ''.join(args)

    def login(self, val_id: int, *groups: str, areas: list = None, replace=False):
        fn = lambda *args: self._hook_set(*(str(it) for it in args), val_id=val_id)
        super().login(fn, *groups, areas=areas, replace=replace)

    # set functions
    def add(self, item):
        self.set.add(item)

    def remove(self, item):
        self.set.remove(item)

    def discard(self, item):
        self.set.discard(item)

    def pop(self):
        return self.set.pop()

    def clear(self):
        self.set.clear()

    def __contains__(self, item):
        return item in self.set

    def __len__(self):
        return len(self.set)

    def __iter__(self):
        return self.set.__iter__()

    def __str__(self):
        return self.set.__str__()

    def __bool__(self):
        return bool(self.set)

    def toset(self):
        return self.set.copy()


class FList(FT):
    """
    Get List data by format
    """

    def __init__(self):
        super().__init__()
        self.list = []

    def _hook_list(self, *args, val_id=None):
        assert val_id is not None, f'val_id should not be None'
        self.list.append(args[val_id])
        return ''.join(args)

    def login(self, val_id: int, *groups: str, areas: list = None, replace=False):
        fn = lambda *args: self._hook_list(*(str(it) for it in args), val_id=val_id)
        super().login(fn, *groups, areas=areas, replace=replace)

    # list functions
    def append(self, item):
        self.list.append(item)

    def remove(self, item):
        self.list.remove(item)

    def pop(self, index=-1):
        return self.list.pop(index)

    def clear(self):
        self.list.clear()

    def __contains__(self, item):
        return item in self.list

    def __len__(self):
        return len(self.list)

    def __getitem__(self, item):
        return self.list[item]

    def __setitem__(self, key, value):
        self.list[key] = value

    def __iter__(self):
        return self.list.__iter__()

    def __str__(self):
        return self.list.__str__()

    def __bool__(self):
        return bool(self.list)

    def tolist(self):
        return self.list.copy()




class FBranch(FT):
    """
    if-else branch
    if true, replace with true_fn(can be None, mean remove it)
    if false, replace with false_fn(can be None, mean remove it)
    """

    def _hook_branch(self, *args, cond_vids=None, cond_fn=None, true_vids=None, true_fn=None, false_vids=None, false_fn=None):
        if cond_vids is None or cond_fn is None:
            raise Exception(f'Condition vid and fn should not be None')
        # condition
        params = []
        for vid in cond_vids:
            if vid >= len(args):
                raise Exception(f'Condition vid {vid} out of range(len = {len(args)})')
            params.append(args[vid])

        if cond_fn(*params):
            # true
            if true_vids is not None and true_fn is not None:
                params = []
                for vid in true_vids:
                    if vid >= len(args):
                        raise Exception(f'True vid {vid} out of range(len = {len(args)})')
                    params.append(args[vid])
                return true_fn(*params)
            else:
                return ''
        else:
            # false
            if false_vids is not None and false_fn is not None:
                params = []
                for vid in false_vids:
                    if vid >= len(args):
                        raise Exception(f'False vid {vid} out of range(len = {len(args)})')
                    params.append(args[vid])
                return false_fn(*params)
            else:
                return ''


    def login(self, cond_vids, cond_fn, true_vids, true_fn, false_vids, false_fn, *groups, areas: list = None):
        """
        :param cond_vids: condition value ids
        :param cond_fn: condition function. return is bool(not a str)
        :param true_vids: true value ids. pass None if no false
        :param true_fn: true function. pass None if no false
        :param false_vids: false value ids. pass None if no false
        :param false_fn: false function. pass None if no false
        :param groups:
        :param areas:
        :return:
        """
        fn = lambda *args: self._hook_branch(*args, cond_vids=cond_vids, cond_fn=cond_fn, true_vids=true_vids, true_fn=true_fn, false_vids=false_vids, false_fn=false_fn)
        super().login(fn, *groups, areas=areas, replace=True, embed=True)


if __name__ == '__main__':
    test = """
        Pip 和 Conda 是 Python 的两大软件包管理工具，它们的官方源在国内访问困难，下载速度非常慢。一般情况下我们使用的都是国内的镜像源，例如清华大学的 TUNA 镜像站、阿里云的镜像站。
    
    但是有些软件包体积非常大，安装的时候从镜像站下载下来仍然需要等待很长时间，如果正巧遇到镜像站负载高峰导致下载速度缓慢，那更是雪上加霜。
    
    为了防止配环境的时候软件包下载等待时间过长，$SINGLE:114$可行的方法就是搭建$SINGLE1:514$本地镜像源，在下载的时候直接从本地镜像源下载，速度能够达到内网带宽。如果是千兆内网，那么理论可以达到 125MB/s，这个速度即使是$SINGLE2:9$ GB 的软件包，也能在半分钟内装好。
    """
    ft = FDict()
    ft.login(0, 1, FT.DOLLAR, FT.VARIABLE, FT.COLON, FT.NUMBER, FT.DOLLAR, areas=[1, 3])
    # print(ft(test))
    # print(ft)

    test = """
    // $if myFunctionA(114):
        1
    // #else
        // $if ctypes.dll(0x44H00F)(DWordData(514)):
            2
        // #else
            3
        // #
        
        // $if 114 > 514:
            4
        // #else
            5
        // #
    // #
    """
    ft = FBranch()
    gs = [
        "//\s*\$if\s+", "[^:]+", ":", "(.|\n)+", "?=//\s*#", "//\s*#else\s*", "(.|\n)+", "?=//\s*#", "//\s*#"
    ]
    ft.login([1], lambda x: eval(x), [3], lambda x: x, [6], lambda x: x, *gs)

    print(ft(test))


# ------------------------------------------------------------------------------------------------------------------------------------- #
# Re new function
def strip(txt:str, __chars:str=None) -> str:
    """
    Strip txt by pattern
    """
    if isinstance(txt, FTMatched):
        txt = txt.matched

    if __chars is None:
        return txt.strip()

    pat_start = f'^[{__chars}]+'
    pat_end = f'[{__chars}]+$'

    start = re.sub(pat_start, '', txt)
    end = re.sub(pat_end, '', start)
    return end

re.strip = strip



from rbpop.win_manager import WinManager, QPop
from rbpop.prefab import *
from rbpop.win_pop import PopWin




if __name__ == '__main__':
    import sys
    import random
    from PyQt5.QtWidgets import QApplication, QLabel
    from PyQt5.Qt import Qt

    class TestWin(PopWin):
        def initUI(self):
            self.resize(360, 120)
            color = f"rgb({random.randint(100, 200)}, {random.randint(150, 200)}, {random.randint(100, 150)})"
            self.setStyleSheet(f"background-color: {color};")
            self.setWindowOpacity(0.8)

            # Add a label, show self
            self.label = QLabel(self)
            self.label.setText(self.__class__.__name__ + '\n' + str(hex(id(self))))
            # Consolas 12, Center
            self.label.setStyleSheet("font-family: Consolas; font-size: 14px;")
            self.label.setAlignment(Qt.AlignCenter)
            # fit widget
            self.label.resize(self.size())

    app = QApplication(sys.argv)
    wm = WinManager()
    wm.add(TestWin(2000))
    wm.add(TestWin(3000))
    wm.add(TestWin(4000))

    sys.exit(app.exec_())











import sys
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QPainter, QColor, QPixmap
# Qpoint
from PyQt5.QtCore import QPoint

def create_triangle_pixel(size: int, color: QColor = QColor("black"), direct="down"):
    assert direct in ["down", "right"], "direct must be 'down' or 'right'"
    # 创建一个 QPixmap 对象，参数为图标大小和是否透明
    pm = QPixmap(size, size)
    pm.fill(Qt.transparent)  # 设置背景透明

    # 创建 QPainter 对象用于绘制
    painter = QPainter(pm)
    painter.setRenderHint(QPainter.Antialiasing)  # 设置反锯齿
    painter.setPen(Qt.NoPen)  # 不绘制边框
    painter.setBrush(color)  # 设置填充颜色

    # 绘制三角形
    if direct == "down":
        points = [QPoint(0, 0), QPoint(size, 0), QPoint(size // 2, size)]  # 三角形的三个顶点
    else:
        points = [QPoint(0, 0), QPoint(0, size), QPoint(size, size // 2)]  # 三角形的三个顶点
    painter.drawPolygon(*points)  # 绘制三角形

    painter.end()  # 完成绘制
    # save it
    pm.save("triangle.png")
    return pm

# 使用图标
from PyQt5.QtWidgets import QApplication, QLabel, QWidget, QVBoxLayout
app = QApplication([])

# 使用函数创建图标
triangle_down_pixel = create_triangle_pixel(16)  # 创建一个大小为 16x16 的向下三角形图标
triangle_right_pixel = create_triangle_pixel(16, direct="right")  # 创建一个大小为 16x16 的向右三角形图标

w = QWidget()
w.setLayout(QVBoxLayout())
w.setWindowTitle("Triangle Icon")
lbl0 = QLabel()
lbl0.setFixedSize(16, 16)
lbl0.setPixmap(triangle_down_pixel)  # 设置图标
w.layout().addWidget(lbl0)
lbl1 = QLabel()
lbl1.setFixedSize(16, 16)
lbl1.setPixmap(triangle_right_pixel)  # 设置图标
w.layout().addWidget(lbl1)
w.setFixedSize(20, 20)
w.show()
sys.exit(app.exec_())




from PyQt5.QtWidgets import QSlider
from PyQt5.QtCore import pyqtSignal, QEvent

class QDoubleSlider(QSlider):
    RATIO = 1000000
    MINIMUM = -2147483647  # use this as real minimum value
    monoChanged = pyqtSignal(float)

    def __init__(self, direction, minimum: float, maximum: float, step: float, parent=None):
        super().__init__(direction, parent)
        self.bias = 0
        self._last_value = self.MINIMUM
        self._on_super_value_changed_flag = False
        self._min, self._max, self._step = minimum, maximum, step
        self.range_max, self.step = self._fit_bounds()
        super().setValue(self.MINIMUM)
        super().valueChanged.connect(self._super_value_changed)

    def _super_value_changed(self, value: int):
        if self._on_super_value_changed_flag:return
        self._on_super_value_changed_flag = True
        # calc a rounded value
        delta = value - self.MINIMUM

        half_step = self.step // 2

        if delta % self.step >= half_step:
            value = self.MINIMUM + int(delta / self.step + 0.5) * self.step
        else:
            value = self.MINIMUM + int(delta / self.step) * self.step

        if value < self.MINIMUM:
            value = self.MINIMUM
        elif value > self.range_max:
            value = self.range_max

        if value == self._last_value:
            self._on_super_value_changed_flag = False
            return

        self._last_value = value
        super().setValue(value)

        rel_value = (value + self.bias) / self.RATIO
        self.monoChanged.emit(rel_value)

        self._on_super_value_changed_flag = False
        return value

    def _fit_bounds(self) -> tuple[int, int]:
        if self._step <= 0:
            raise ValueError("Step cannot be less than or equal to 0.")
        if self._min >= self._max:
            raise ValueError("Minimum cannot be greater than or equal to maximum.")

        minimum = int(self._min * self.RATIO)
        maximum = int(self._max * self.RATIO)
        step = int(self._step * self.RATIO)
        delta = minimum - self.MINIMUM

        self.bias = delta

        super().setMinimum(self.MINIMUM)
        super().setMaximum(maximum - delta)
        super().setSingleStep(step)
        super().setPageStep(step)
        return (maximum - delta, step)

    def value(self) -> float:
        return (super().value() + self.bias) / self.RATIO

    def setValue(self, value: float):
        super().setValue(int(value * self.RATIO - self.bias))
        self.monoChanged.emit(value)

    def setMinimum(self, minimum: float):
        self._min = minimum
        self.range_max, self.step = self._fit_bounds()

    def setMaximum(self, maximum: float):
        self._max = maximum
        self.range_max, self.step = self._fit_bounds()

    def setSingleStep(self, step: float):
        self._step = step
        self.range_max, self.step = self._fit_bounds()

    def setPageStep(self, a0):
        raise NotImplementedError("Page step is not supported in QDoubleSlider.")


class QIntSlider(QSlider):
    monoChanged = pyqtSignal(int)

    def __init__(self, direction, minimum: int, maximum: int, step: int=1, parent=None):
        super().__init__(direction, parent)
        self._last_value = minimum
        self._on_super_value_changed_flag = False
        super().setMinimum(minimum)
        super().setMaximum(maximum)
        super().setSingleStep(step)
        super().setPageStep(step)
        super().setValue(minimum)
        super().valueChanged.connect(self._super_value_changed)

    def _super_value_changed(self, value: int):
        if self._on_super_value_changed_flag:return
        self._on_super_value_changed_flag = True
        # calc a rounded value
        minimum = self.minimum()
        maximum = self.maximum()
        delta = value - minimum
        step = self.singleStep()

        half_step = step // 2

        if delta % step >= half_step:
            value = minimum + int(delta / step + 0.5) * step
        else:
            value = minimum + int(delta / step) * step

        value = int(value + .5)

        if value < minimum:
            value = minimum
        elif value > maximum:
            value = maximum

        if value == self._last_value:
            self._on_super_value_changed_flag = False
            return

        self._last_value = value
        super().setValue(value)

        self.monoChanged.emit(value)

        self._on_super_value_changed_flag = False
        return value

    def setSingleStep(self, step: int):
        super().setSingleStep(step)
        super().setPageStep(step)

    def setPageStep(self, a0):
        raise NotImplementedError("Page step is not supported in QDoubleSlider.")

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout
    from PyQt5.QtCore import Qt

    app = QApplication([])
    window = QWidget()
    layout = QVBoxLayout()

    double_slider = QDoubleSlider(Qt.Horizontal,0.0, 100.0, 5.0)
    double_slider.monoChanged.connect(lambda value: print(f"Value: {value:.6f}"))
    layout.addWidget(double_slider)

    int_slider = QIntSlider(Qt.Horizontal,0, 100, 5)
    int_slider.monoChanged.connect(lambda value: print(f"Value: {value}"))
    layout.addWidget(int_slider)

    window.setLayout(layout)
    window.show()
    app.exec_()


import ast

from reft.ft import *
try:
    from PyQt5.QtWidgets import QWidget
    from mono_ui import QMonoWidget
except ImportError:
    pass

__doc__ = """
mono.py - based on reft
用来处理一种类似unity mono的attributes的标注语言嵌入
*允许嵌入到任何代码中，用于控制目标代码的最终内容
*可以支持python eval exec、if branch等
*使用Attributes修饰Variable，可以提供创建Inspector的PyQt5 Widget的API

ESO eval EEO

VSO express VEO

XSO
ASO attribute1|attribute2... AEO 
exec 
XEO

IFSO eval
...
EFSO eval
...
ELSO
...
IFEO

* 不支持LOOP
"""

_DEBUG_MODE = False


class FTMonoInfiniteExecLoop(Exception):
    pass


class FTMonoAttrToMultiExecs(Exception):
    pass

class FTMonoEvalError(Exception):
    def __init__(self, raw:FTMatched, env:dict):
        txt = f" Failed to eval: '{raw}' at lineno {raw.lineno} with environ:\n\t{env}"
        super().__init__(txt)

class MonoAttr:
    def __init__(self, name, value, *attrs, lineno=None):
        self._name = name
        self._value = value
        self._attrs = attrs
        self._lineno = lineno

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._value = v

    @property
    def attrs(self):
        return self._attrs

    @property
    def lineno(self):
        return self._lineno if self._lineno is not None else -1

    def iterall(self, func):
        """
        遍历更新所有的属性
        """
        _news = []
        for i, attr in enumerate(self._attrs):
            try:
                _news.append(func(attr))
            except Exception as e:
                raise ValueError(f"MonoAttr.iterall:Failed to update attr {attr} with error: {e}")
        self._attrs = tuple(_news)

    def __str__(self):
        return f"({self._name}={self._value})[{'|'.join(self._attrs)}]"

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self._name == other._name

    def __hash__(self):
        return hash(self._name)

    def __getitem__(self, item):
        return self._attrs[item]

    def __iter__(self):
        return iter(self._attrs)





def api_parse_python_single_line_vardef(s: str, env: dict = None) -> tuple[str, any] | str | None:
    """
    解析python单行变量定义
    * 1. 只支持单行
    * 2. 只支持变量定义(包括赋值)
    * 3. 使用提供的env字典来计算变量值
    * 4. 返回变量名和变量值(如果有)

    * env传入None时，不会计算变量值，只返回变量名
    """
    try:
        # 解析字符串为AST
        tree = ast.parse(s, mode='exec')
    except SyntaxError:
        # 如果解析失败，返回None
        return None

    # 遍历AST节点
    for node in ast.walk(tree):
        # 检查是否是赋值语句
        if isinstance(node, ast.Assign):
            # 检查是否只有一个目标（即一个变量）
            if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
                var_name = node.targets[0].id
                if env is None:
                    return var_name
                # 将AST节点转换为可执行代码
                code = compile(ast.Expression(node.value), filename="<ast>", mode="eval")
                # 使用env字典计算变量值
                var_value = eval(code, env)
                return var_name, var_value
    return None


def api_ftmono_handle_execs(execs: list[FTMatched], env: dict, depth=FT_DEPTH) -> dict:
    """
    处理execs
    * 1. 轮流处理execs, 记录结果没有报错的index
    * 2. 如果某轮处理全部报错，那么停止处理
    """
    _passed = []  # 记录总的通过的index
    _aimnum = len(execs)  # 目标处理次数
    _left = depth  # 剩余的处理次数

    # empty check
    if not _aimnum: return {}

    # loop
    while _left > 0:
        _left -= 1
        _epoch_passed = []  # 记录本轮通过的index
        _epoch_errors = []  # 记录本轮错误的内容
        for i, s in enumerate(execs):
            ftm: FTMatched = s
            lineno, content = ftm.lineno, str(ftm)
            if i in _passed: continue  # skip passed items
            try:
                exec(content, env)
                _epoch_passed.append(i)
            except Exception as e:
                _epoch_errors.append((i, e))
        if not _epoch_passed:
            _lefts = [f"lineno:{lineno}" for i in range(len(execs)) if i not in _passed]
            _erros = [str(i) + ": " + str(e) for i, e in _epoch_errors]
            _envs = {k: v for k, v in env.items() if not k.startswith("__")}
            raise FTMonoInfiniteExecLoop("Infinite Exec Loop Detected" +
                                         f"\n\tenvironments: \n\t\t{_envs}" +
                                         f"\n\tlefts: \n\t\t{_lefts}" +
                                         f"\n\terrors: \n\t\t{_erros}")
        _passed.extend(_epoch_passed)
        if len(_passed) == _aimnum:
            break
    return env


def api_parse_ifbr_single(s: str, IFSO: str, EFSO: str, ELSO: str, IFEO: str) ->  list[tuple[str, str | None, str]]:
    """
    解析单个的if branch, 主要用于提取ifbr single的类别，EVAL(如果有的话)，body
    比如:
    'IFSO EVAL
    ...'
    * 特别的，允许接受连续的多个elif语句
    """
    # check 'IFSO EVAL \n BODY'
    gs = [
        f"{FT.BNN}*", IFSO, f"{FT.BNN}*", ".+", f"{FT.ANY_CHAR}+"
    ]
    vids = [1, 3, 4]
    res = FT.Extract(s, *gs, areas=vids)
    if res:
        return res

    # check 'EFSO EVAL \n BODY'
    gs = [
        f"{FT.BNN}*", EFSO, f"{FT.BNN}*", ".+", f"{FT.ANY_CHAR}+", FT.WF(True, EFSO)
    ]
    vids = [1, 3, 4]
    _match = re.search(EFSO, str(s))
    if _match is not None:  # ELSO 分支
        res = FT.Extract(str(s) + _match.group(0), *gs, areas=vids)  # 由于使用了FT.WrapForward，所以需要补充一个EFSO
        if res:
            return res

    # check 'XSO BODY'
    gs = [
        f"{FT.BNN}*", ELSO, f"{FT.BNN}*", f"{FT.ANY_CHAR}+"
    ]
    vids = [1, 3]
    tmp = FT.Extract(s, *gs, areas=vids)
    res = []
    for a, b in tmp:
        res.append((a, None, b))
    # ret
    if res:
        return res
    return []

def api_select_ifbr_branch_content(tri_tuples:list[tuple[FTMatched]], env) -> tuple[int, str|None]:
    """
    根据env选择目标ifbr组合的对应body content
    返回None表示未匹配到，返回tuple表示匹配到的内容的表达式索引和内容
    """
    if not tri_tuples:
        return ""

    # clean 中段
    _new = []
    for a, b, c in tri_tuples:
        if b is not None:
            b = re.strip(b, "\s")
        _new.append((a, b, c))
    tri_tuples = _new

    # build target
    TARGET = "api_build_ifbr_exec_target"
    txt = f"{TARGET}=None\n"

    if tri_tuples[0][1] is None:
        raise ValueError(f"Invalid ifbr tri_tuples: {tri_tuples} without first if.")
    txt += f"if {tri_tuples[0][1]}:\n\t{TARGET}= 0\n"
    for i, (XSO, EVAL, BODY) in enumerate(tri_tuples[1:]):
        if EVAL:
            txt += f"elif {EVAL}:\n\t{TARGET}= {i+1}\n"
        else:
            txt += f"else:\n\t{TARGET}= {i+1}\n"

    # try to exec
    envc = env.copy()
    try:
        exec(txt, envc)
    except Exception as e:
        raise ValueError(f"Exec target built code with error: {e}\n\ntxt:\n{txt}")

    index = envc[TARGET]
    if index is None:
        return -1, ""
    _res = tri_tuples[envc[TARGET]][2]
    return envc[TARGET], _res


class Mono:
    @staticmethod
    def _api_count_newline(*strs):
        _ = 0
        for s in strs:
            _ += str(s).count('\n')
        return _

    def __init__(self, ESO, EEO, VSO, VEO,
                 XSO=None, XEO=None, ASO=None, AEO=None,
                 IFSO=None, EFSO=None, ELSO=None, IFEO=None,
                 *, COMMENT=';', CSO="/\*", CEO="\*/", ENV: str = None):
        """
        '
        ESO exec EEO  ; 该语句必须在一行内

        VSO express VEO  ; 嵌入表达式结果到目标

        '
        XSO
        exec
        XEO

        '
        ASO attribute AEO  ; 只能用在eval语句前，并且只作用于渲染eval中定义的变量
        ESO eval EEO

        '   ; 根据表达式控制目标代码的分支
        IFSO eval
        ...
        EFSO eval
        ...
        ELSO ;else
        ...
        IFEO

        * ES?可以和XS?一致。程序会根据包含内容是否有多行自动推断
        * 如果不指定XSO XEO，那么会默认和ESO EEO一致
        * 如果不指定ASO AEO，那么会忽略attribute
        * 如果不指定IFSO EFSO ELSO IFEO，那么会忽略if else
        * 如果不指定LPSO LPBK LPCT LPEO，那么会忽略loop

        * 默认使用;作为注释符号，可以通过comment参数修改
        * 默认使用/\* \*/作为注释符号，可以通过CSO CEO参数修改

        * env: 用于在构造eval和exec用的虚拟环境时执行的代码。一般可以指定为几条import语句
        """
        self._ESO = ESO
        self._EEO = EEO
        self._VSO = VSO
        self._VEO = VEO
        self._XSO = XSO or ESO
        self._XEO = XEO or EEO
        self._ASO = ASO or None
        self._AEO = AEO or None
        self._IFSO = IFSO or None
        self._EFSO = EFSO or None
        self._ELSO = ELSO or None
        self._IFEO = IFEO or None
        self._COMMENT = COMMENT
        self._CSO = CSO
        self._CEO = CEO
        self._ENV = ENV
        assert isinstance(ESO, str), f"Mono's Exec-Line Start Operator must be a string, but got {ESO}"
        assert isinstance(EEO, str), f"Mono's Exec-Line End Operator must be a string, but got {EEO}"
        assert isinstance(VSO, str), f"Mono's eval-Value Start Operator must be a string, but got {VSO}"
        assert isinstance(VEO, str), f"Mono's eval-Value End Operator must be a string, but got {VEO}"
        assert isinstance(XSO or ESO, str), f"Mono's Exec-MultipleLines Start Operator must be a string, but got {XSO}"
        assert isinstance(XEO or EEO, str), f"Mono's Exec-MultipleLines End Operator must be a string, but got {XEO}"
        assert ASO is None or isinstance(ASO, str), f"Mono's Attribute Start Operator must be a string, but got {ASO}"
        assert AEO is None or isinstance(AEO, str), f"Mono's Attribute End Operator must be a string, but got {AEO}"
        assert IFSO is None or isinstance(IFSO, str), f"Mono's If-Branch Start Operator must be a string, but got {IFSO}"
        assert EFSO is None or isinstance(EFSO, str), f"Mono's Elif-Branch Start Operator must be a string, but got {EFSO}"
        assert ELSO is None or isinstance(ELSO, str), f"Mono's Else-Branch Start Operator must be a string, but got {ELSO}"
        assert IFEO is None or isinstance(IFEO, str), f"Mono's If-Branch End Operator must be a string, but got {IFEO}"
        assert isinstance(COMMENT, str), f"Mono's Comment Operator must be a string, but got {COMMENT}"
        assert isinstance(CSO, str), f"Mono's Comment Start Operator must be a string, but got {CSO}"
        assert isinstance(CEO, str), f"Mono's Comment End Operator must be a string, but got {CEO}"
        assert isinstance(ENV, str), f"Mono's Env must be a string, but got {ENV}"
        if self.with_if_branch:
            if '\n' in IFSO or '\n' in EFSO or '\n' in ELSO or '\n' in IFEO:
                raise ValueError(f"If-Branch Operators can not contain newline. Got'{IFSO}' '{EFSO}' '{ELSO}' '{IFEO}'")

        # ------------------------------
        self._dgs = self._build_gs()
        self._ft = self._build_ft()
        self._sft = self._build_sft()
        self._ift = self._build_ifbr_ft()
        self._env = self._build_env()
        self._overenv = {}  # 用于指定覆盖内容的env
        self._execs = []  # 用于存储运行时exec-multilines  # list[FTMatched]
        self._exevs = []  # 用于存储运行时exec-line的内容  # list[FTMatched]
        self._attrs = []  # 用于存储运行时attr的内容  # list[FTMatched]
        self._binding = []  # 用于存储attr和exec的绑定关系  # list[MonoAttr]
        self._brchs = []  # 用于存储运行时branch的内容  # list[(is_ifbr:bool, cond/iter :FTMatched, body:FTMatched)]  # true: if false: loop


    @property
    def dgs(self):
        return self._dgs

    @property
    def ft(self):
        return self._ft

    @property
    def monos(self):
        return self._binding.copy()

    @property
    def dict(self):
        initial = self._build_env()
        current = self.env
        return {k: v for k, v in current.items() if k not in initial and not k.startswith("__")}

    @property
    def types(self):
        return {k: type(v) for k, v in self.dict.items()}

    @property
    def keys(self):
        dfs = self.dict
        return list(dfs.keys())

    @property
    def env(self):
        return self._env.copy()

    @property
    def envs(self):
        """
        获取一个用于展示的env
        """
        return {k: v for k, v in self._env.items() if not k.startswith("__")}

    @property
    def with_attrubute(self):
        return self._ASO is not None and self._AEO is not None

    @property
    def with_if_branch(self):
        return self._IFSO is not None and self._IFEO is not None and self._EFSO is not None and self._ELSO is not None

    def _build_gs(self):
        """
        根据设定值创建Groups
        """
        # 捕获 ESO express_line EEO  的格式
        epgs, epids = [f"{self._ESO}{FT.BNN}*", f".+", f"?=({FT.BNN}*{self._EEO})", f"{FT.BNN}*{self._EEO}"], [0, 1, 3]
        # 捕获 XSO express_line XEO  的格式
        xpgs, xvids = [f"{self._XSO}{FT.BNN}*", f"{FT.ANY_CHAR}+", f"?=({FT.BNN}*{self._XEO})", f"{FT.BNN}*{self._XEO}"], [0, 1, 3]
        # 捕获 VSO express_line VEO  的格式
        vlgs, vlids = [f"{self._VSO}{FT.BNN}*", f".+", f"?=({FT.BNN}*{self._VEO})", f"{FT.BNN}*{self._VEO}"], [0, 1, 3]
        # 捕获 ASO attribute AEO  的格式
        apgs, aaids = [f"{self._ASO}{FT.BNN}*", f".+", f"?=({FT.BNN}*{self._AEO})", f"{FT.BNN}*{self._AEO}"], [0, 1, 3]
        # 捕获 BRSO express_lines BREO  的格式(cvid:1)
        brgs, brids = [f"{self._IFSO}{FT.BNN}*", f"{FT.ANY_CHAR}+", f"?=({FT.BNN}*{self._IFEO})", f"{FT.BNN}*{self._IFEO}"], [0, 1, 3]
        # ifbr 检测
        ifgs = [
            f"{FT.BNN}*{self._IFSO}{FT.BNN}*.*((.|\n)+?)({FT.WF(True, f'({self._EFSO})|({self._ELSO})|({self._IFEO})')})",
            f"({FT.BNN}*{self._EFSO}{FT.BNN}*.*((.|\n)+?)({FT.WF(True, f'({self._EFSO})|({self._ELSO})|({self._IFEO})')}))*",
            f"({FT.BNN}*{self._ELSO}{FT.BNN}*((.|\n)+?)({FT.WF(True, self._IFEO)}))?",
            f"{FT.BNN}*{self._IFEO}{FT.BNN}*"
        ]
        ifids = [0, 1, 2, 3]
        efs = f"{FT.BNN}*{self._EFSO}{FT.BNN}*.*((.|\n)+?)({FT.WF(True, self._EFSO)}))"
        eids = [0]

        # 创建移除comment的group
        rmgs0, rmgs1 = [self._COMMENT, ".*"], [self._CSO, f"{FT.ANY_CHAR}+", f"{FT.WF(True, self._CEO)}", self._CEO]

        return {
            'exec': [epgs, epids],
            'execs': [xpgs, xvids],
            'eval': [vlgs, vlids],
            'attr': [apgs, aaids],
            'brch': [brgs, brids],
            'ifbr': [ifgs, ifids],  # 'ifbr': 'if branch
            'rmgs': [rmgs0, rmgs1],
            '_efs': [efs, eids]
        }

    def _build_ft(self):
        """
        根据设定值创建FT
        """
        ft = FT()
        ft.login(lambda *args: "", *self._dgs['rmgs'][0], areas=[1])
        ft.login(self._hook_comments, *self._dgs['rmgs'][1], areas=[0, 1, 3])
        ft.login(self.clear)  # Pass-In mode login. 用于重置状态
        ft.login(self._exec_line_hook, *self._dgs['exec'][0], areas=self._dgs['exec'][1])
        ft.login(self._exec_multilines_hook, *self._dgs['execs'][0], areas=self._dgs['execs'][1])
        ft.login(self._handle_execs)  # Pass-In mode login. 用于处理exec
        ft.login(self._update_env)  # Pass-In mode login. 用于处理后续之前更新env
        if self.with_if_branch or self.with_loop_branch:
            ft.login(self._brch_hook, *self._dgs['brch'][0], areas=self._dgs['brch'][1], embed=True)
        ft.login(self._eval_hook, *self._dgs['eval'][0], areas=self._dgs['eval'][1])
        if self.with_attrubute:
            ft.login(self._attr_hook, *self._dgs['attr'][0], areas=self._dgs['attr'][1])
            ft.login(self._handle_attr)  # Pass-In mode login. 用于处理attr
        # clear multiple lines
        ft.login(lambda *args: '\n', "(\n[ \t]*){2,}\n")
        ft.login(lambda *args: '', "^(\n[ \t]*)+")  # clear start

        return ft

    def _build_sft(self):
        """
        创建Slave FT
        """
        sft = FT()
        sft.login(lambda *args: "", *self._dgs['rmgs'][0], areas=[1])
        sft.login(self._hook_comments, *self._dgs['rmgs'][1], areas=[0, 1, 3])
        if self.with_if_branch or self.with_loop_branch:
            sft.login(self._brch_hook, *self._dgs['brch'][0], areas=self._dgs['brch'][1], embed=True)
        sft.login(self._eval_hook, *self._dgs['eval'][0], areas=self._dgs['eval'][1])
        # clear multiple lines
        sft.login(lambda *args: '\n', "(\n[ \t]*){2,}\n")
        sft.login(lambda *args: '', "^(\n[ \t]*)+")  # clear start

        return sft



    def _build_ifbr_ft(self):
        """
        创建ifbr_ft
        """
        ft = FT()
        ft.login(self._ifbr_hook, *self._dgs['ifbr'][0], areas=self._dgs['ifbr'][1])
        return ft

    def _build_env(self):
        """
        创建虚拟环境
        """
        env = {}
        if self._ENV:
            exec(self._ENV, env)
        return env

    def _update_env(self):
        """
        更新env. 用于在处理eval replace之前更新env
        """
        self._env.update(self._overenv)

    def clear(self):
        """
        清空执行环境
        """
        self._execs.clear()
        self._exevs.clear()
        self._attrs.clear()
        self._binding.clear()
        self._brchs.clear()
        self._env = self._build_env()

    @staticmethod
    def _triple_wrap_newlines(_0, raw_s, rep_s, _1, *, rep=True):
        """
        对经典的三个group进行补充换行符的处理。
        * 有助于你保持lineno
        * rep: 是否使用rep_s插入到目标输出中
        """
        raws_cnt, reps_cnt = Mono._api_count_newline(raw_s), Mono._api_count_newline(rep_s)
        _0_cnt, _1_cnt = Mono._api_count_newline(_0), Mono._api_count_newline(_1)
        if not rep:
            assert raws_cnt == reps_cnt, f"raws_cnt:{raws_cnt} reps_cnt:{reps_cnt}\nThis is a reft-lib error. Not user's\ndetail: Should pass in same raw_s and rep_s if use rep=False."
            return (_0_cnt + reps_cnt + _1_cnt) * '\n'
        else:
            if reps_cnt > raws_cnt:  # 目标替换的内容比原内容的换行符更多
                added = reps_cnt - raws_cnt
                if added > _1_cnt:  # _0是不能被优化的，它会影响rep_s的lineno，只能登记'膨胀信息'到current
                    _vno, _bno = _1.lineno, added - _1_cnt
                    _1_cnt = 0
                else:
                    _1_cnt -= added
            return _0_cnt * '\n' + str(rep_s) + _1_cnt * '\n'

    def _hook_comments(self, _0, s: FTMatched, _1):
        """
        处理comments
        """
        return self._triple_wrap_newlines(_0, s, s, _1, rep=False)

    def _exec_line_hook(self, _0, s: FTMatched, _1):
        """
        处理eval
        """
        if _DEBUG_MODE: print("exec", s)
        self._exevs.append(s)
        return self._triple_wrap_newlines(_0, s, s, _1, rep=False)

    def _exec_multilines_hook(self, _0, s: FTMatched, _1):
        """
        处理exec
        """
        if _DEBUG_MODE: print("execs", s)
        self._execs.append(s)
        return self._triple_wrap_newlines(_0, s, s, _1, rep=False)

    def _handle_execs(self, depth=FT_DEPTH):
        self._env = api_ftmono_handle_execs(self._exevs + self._execs, self._env, depth=depth)
        if _DEBUG_MODE: print("execs passed, env=", self.envs)

    def _brch_hook(self, head: FTMatched, s: FTMatched, end: FTMatched):
        """
        处理brch
        """
        # is_ifbr = re.fullmatch(self._IFSO, str(head)) is not None
        _s = str(head) + str(s) + str(end)

        # May be you can do somethings here to support loop

        # fill \n
        rep = self._ift.handle(_s)
        _group0 = s.group0  # s所在的re.Match对象的group(0)
        # _before_idx = _group0.
        return rep

    def _ifbr_hook(self, if_expr, ef_expr, el_expr, ed_expr):
        """
        处理ifbr
        """
        # build exprs
        exprs: list[tuple[str, str, str]] = []  # [(IF, EVAL, BODY)]
        exprs = api_parse_ifbr_single(if_expr, self._IFSO, self._EFSO, self._ELSO, self._IFEO)
        exprs += api_parse_ifbr_single(ef_expr, self._IFSO, self._EFSO, self._ELSO, self._IFEO)
        exprs += api_parse_ifbr_single(el_expr, self._IFSO, self._EFSO, self._ELSO, self._IFEO)
        if _DEBUG_MODE: print("ifbr", *exprs, sep='\n')

        idx, res = api_select_ifbr_branch_content(exprs, self._env)
        if _DEBUG_MODE: print("ifbr passed, res=", res)

        # rebuild multiples \n
        if idx ==-1:
            count = 0
            for i in range(len(exprs)):
                count += str(exprs[i][1]).count('\n') + str(exprs[i][2]).count('\n')
            res = count * '\n'
        else:
            # 前置已知: IFSO EFSO ELSO IFEO 不包含\n; \n不会被areas外的group捕获
            before_count, after_count = 0, 0
            for i in range(idx):
                before_count += self._api_count_newline(*exprs[i])
            for i in range(idx + 1, len(exprs)):
                after_count += self._api_count_newline(*exprs[i])
            res = before_count * '\n' + res + after_count * '\n'

        return res + self._api_count_newline(ed_expr) * '\n'

    def _eval_hook(self, _0, s: FTMatched, _1):
        """
        处理eval
        """
        if _DEBUG_MODE: print("eval", s)

        # handle eval
        try:
            _ev = eval(str(s), self._env)
        except Exception as e:
            raise FTMonoEvalError(s, self.envs)
        return self._triple_wrap_newlines(_0, s, _ev, _1, rep=True)

    def _attr_hook(self, _0, s: FTMatched, _1):
        """
        处理attr
        """
        if _DEBUG_MODE: print("attr", s)

        self._attrs.append(s)
        return self._triple_wrap_newlines(_0, s, s, _1, rep=False)

    def _handle_attr(self):
        """
        处理attr(self._attrs)
        * 1. 每个s:FTMatched都有一个lineno的属性。把一个或多个attr绑定到一个exec-line上。
        * 2. exec-multilines可以打断attr的绑定, 此时抛出异常FTMonoAttrToMultiExecs
        * self._execs multilines; self._exevs line
        * 均为list[FTMatched]
        * 使用api_parse_python_single_line_vardef获取变量名
        :return: dict  # {var_name of _exevs: [s of _attr]}

        例如:
        // [Type(BOOL)]
        // [READONLY]
        //> add_en=False

        exevs: ["add_en=False"], attrs: ["Type(BOOL)", "READONLY"] execs: []
        return {"add_en": ["Type(BOOL)", "READONLY"]}
        """
        _items = []
        TYPE_EXEV, TYPE_EXECS, TYPE_ATTRS = 0, 1, 2
        for s in self._exevs:
            s._MONO_TYPE = TYPE_EXEV
            _items.append(s)
        for s in self._execs:
            s._MONO_TYPE = TYPE_EXECS
            _items.append(s)
        for s in self._attrs:
            s._MONO_TYPE = TYPE_ATTRS
            _items.append(s)

        _items.sort(key=lambda x: x.lineno)  # 按照行号排序

        _res = []
        _acc_attrs = []  # 用于存储当前尚未被打断的连续的attr列表
        for i, s in enumerate(_items):
            if s._MONO_TYPE == TYPE_ATTRS:
                # 将a|b|c ... 转换为 [a, b, c, ...]:list
                s_lst = [i.strip() for i in str(s).split('|') if i.strip()]
                _acc_attrs.extend(s_lst)
            elif s._MONO_TYPE == TYPE_EXEV:
                if _acc_attrs:
                    var_name, var_value = api_parse_python_single_line_vardef(str(s), self.env)  # 获取变量名和当前值(*如果没有指定覆盖参数，那么就是默认值)
                    if var_name:
                        _res.append(MonoAttr(var_name, var_value, *_acc_attrs, lineno=s.lineno))
                        _acc_attrs = []  # 重置属性列表
            elif s._MONO_TYPE == TYPE_EXECS and _acc_attrs:
                raise FTMonoAttrToMultiExecs("Attrubites can not bind to multiple-execs.")
            # else:  # 不应该出现这种情况
            #     raise

        if _DEBUG_MODE: print("attrs passed, res=", _res)
        self._binding = _res

    def handle(self, s, **params):
        """
        处理目标代码，并返回处理后的结果。
        * 使用params来传递覆盖的参数
        """
        self._overenv = params
        res = self._ft.handle(s)
        self._overenv = {}
        return res

    def build(self, s) -> str | None:
        """
        解析目标代码，并将有关结果存储在self中. 解析失败时返回错误str
        """
        try:
            self._ft.handle(s)
        except Exception as e:
            return str(e)

    def slaveHandle(self, s) -> str:
        """
        根据已有的环境对目标代码进行处理。
        该处理不会改变env，意味着只有eval和ifbr可以被处理
        """
        res = self._sft.handle(s)
        return res




if __name__ == '__main__':
    mono = Mono(
        "//\s*>", "\n",
        "\$", "\$",
        None, "//\s*<",
        "//\s*\[", "\]",
        "//\s*\?", '//\s*:\?', "//\s*:", "//\s*\$",
        COMMENT=r";#", CSO=r"/*#", CEO=r"\*/",
        ENV=r"import math"
    )

    with open('counter.v5.v', 'r', encoding="utf-8") as f:
        test = f.read()

    with open('saved.v', 'w', encoding="utf-8") as f:
        f.write(mono.handle(test, WIDTH=10, add_en=True))

    print(mono.monos)



"""
支持以下attrs-inspector:
* attr可以使用动态参数，传入VSO xxx VEO来指定动态参数
* 可以被ifbr语句控制

如果名称前有**，表示可以多次出现在一个vardef前

Type(X)  ;# 允许指定基础类型
    X can be one of the following types:
    - int   (QSpinBox)
    - float (QDoubleSpinBox)
    - str   (QLineEdit)
    - bool  (QCheckBox)
    * None会被解释为str类型的""
Range(X[, Y[, S) ;# 允许指定范围
    - X: 当只有一个参数时，表示最大值；否则表示最小值
    - Y: 最大值
    - S: 步长 必须>0
    * 这实质上会创建一个python range对象，所以范围的取值是左闭右开的 [X, Y)
    * 当描述符与Type(int)结合时，提升QSpinBox为QSlider，且步长默认值为1
    * 当描述符与Type(float)结合时，提升QDoubleSpinBox为QSlider，且步长默认值为0.01
    * 当描述符与Type(int|float)结合并且S的类型不匹配时，会抛出异常
    * 当描述符与Type(str|bool)结合时，会抛出异常
Enum(X, Y, Z, ...) ;# 强制要求输入必须为枚举值
    * X, Y, Z, ...: 枚举值
    * 当描述符与Type(int|float|str)结合时，提升QXXX为QComboBox
    * QComboBox和QSlider互相冲突，最终会以QComboBox为准
    * 当描述符与Type(bool)结合时，会抛出异常
Color(fg[, bg) ;# 设置整个控件的全局前景色和全局背景色
    * fg: 前景色
    * bg: 背景色
    * 支持 #RRGGBB, rgb(r, g, b) 格式
Tooltip(X) ;# 设置控件的提示信息
Label(X) ;# 设置别名标签
    * 如果没有设置别名标签，会使用变量名作为标签的文本
**Group([X[,color) ;# 接下来的控件会被分组到Group:X中
    * X: 分组的标题
    * color: 分组的前景色
    * 当没有指定X时，接下来的控件会重新回到上一级分组
    * 一个Group是可以被折叠起来的
**Header(X[, color) ;# 在此处插入一个标题，例如 X:
    * X: 标题
    * color: 前景色
**Title(X[, color) ;# 在此处插入一个标题，例如 X:
    * X: 标题
    * color: 前景色
    * Title是一个比Header拥有更大字体和特殊阴影效果的标题
**Separator() ;# 在此处插入一个分隔符
**Space([x) ;# 在此处插入一个空白控件，高度为x.
    * x: 高度
    * 当没有指定x时，高度为标准间距
"""
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QSpinBox, QDoubleSpinBox, QLineEdit, QCheckBox, QComboBox, QPushButton, QApplication, QFrame, QScrollArea
from PyQt5.QtCore import Qt, QPropertyAnimation, QPoint, QRect, pyqtSignal
from PyQt5.QtGui import QFont, QPainter, QPen, QColor, QBrush, QPixmap, QIcon
from reft._mono_qsliders import QDoubleSlider, QIntSlider
from reft._mono_qshadows import QShadowLabel
from reft.mono import Mono, MonoAttr
from reft._rangef import rangef, find_closest
import warnings
import types
import sys
import re
warnings.filterwarnings("ignore", category=DeprecationWarning)
BUILTIN_TYPES = (
    int, float, complex, str, bool, bytes, bytearray, memoryview,
    tuple, list, dict, set, frozenset,
    types.FunctionType, types.BuiltinFunctionType, types.MethodType,
    types.LambdaType, types.CodeType, types.ModuleType,
    types.TracebackType, types.FrameType, types.GeneratorType,
    types.CoroutineType
)


def str_crop(s, size=200):
    s = str(s)
    if len(s) > size:
        return s[:size] + '...'
    return s

def _is_builtin_instance(inst) -> bool:
    """
    检查对象 inst 是否是 Python 的内置类型实例。

    参数:
    inst -- 要检查的对象

    返回:
    如果是内置类型实例，则返回 True，否则返回 False
    """
    # 定义一个元组，包含所有内置类型的类型

    # 检查 inst 的类型是否在这个元组中
    return isinstance(inst, BUILTIN_TYPES)

def api_hide_layout_widgets(layout):
    for i in range(layout.count()):
        item = layout.itemAt(i)
        widget = item.widget()
        if widget:
            widget.hide()
        else:
            if item.layout():
                api_hide_layout_widgets(item.layout())

def api_show_layout_widgets(layout):
    for i in range(layout.count()):
        item = layout.itemAt(i)
        widget = item.widget()
        if widget:
            widget.show()
        else:
            if item.layout():
                api_show_layout_widgets(item.layout())

def api_create_triangle_pixel(size: int, color: QColor = QColor("black"), direct="down"):
    assert direct in ["down", "right"], "direct must be 'down' or 'right'"
    # 创建一个 QPixmap 对象，参数为图标大小和是否透明
    pm = QPixmap(size, size)
    pm.fill(Qt.transparent)  # 设置背景透明

    # 创建 QPainter 对象用于绘制
    painter = QPainter(pm)
    painter.setRenderHint(QPainter.Antialiasing)  # 设置反锯齿
    painter.setPen(Qt.NoPen)  # 不绘制边框
    painter.setBrush(color)  # 设置填充颜色

    # 绘制三角形
    if direct == "down":
        points = [QPoint(0, 0), QPoint(size, 0), QPoint(size // 2, size)]  # 三角形的三个顶点
    else:
        points = [QPoint(0, 0), QPoint(0, size), QPoint(size, size // 2)]  # 三角形的三个顶点
    painter.drawPolygon(*points)  # 绘制三角形

    painter.end()  # 完成绘制
    # save it
    pm.save("triangle.png")
    return pm


def api_merge_colors(*qc):
    # 初始化颜色累加器
    red_sum, green_sum, blue_sum, alpha_sum = 0, 0, 0, 0
    count = 0       # 权重和

    # 累加所有颜色的 RGB 和 Alpha 值
    for color in qc:
        if isinstance(color, tuple):
            weight, color = color
        else:
            weight = 1
        red_sum += color.red() * weight
        green_sum += color.green() * weight
        blue_sum += color.blue() * weight
        alpha_sum += color.alpha() * weight
        count += weight

    # 计算平均值
    if count > 0:
        red_avg = red_sum // count
        green_avg = green_sum // count
        blue_avg = blue_sum // count
        alpha_avg = alpha_sum // count
    else:
        # 如果没有传入任何颜色，返回一个默认的透明颜色
        return QColor(0, 0, 0, 0)

    # 创建并返回新的 QColor 对象
    return QColor(red_avg, green_avg, blue_avg, alpha_avg)

class FTMonoaRuntimeEvalError(Exception):
    def __init__(self, monoa, msg):
        txt = f"\n\t{str(msg)}\n\n\tLineno:{monoa.lineno}, Target:\n\t\t{monoa}"
        super().__init__(txt)

class FTQMonoWidgetRecvObjectError(Exception):
    pass

class FTMonoaRuntimeUnexpectedColor(FTMonoaRuntimeEvalError):
    pass

class FTMonoaRuntimeDismatchedEnumType(FTMonoaRuntimeEvalError):
    pass

class FTMonoaRuntimeDismatchedRangeType(FTMonoaRuntimeEvalError):
    pass


def _default_color(FG_COLOR, FG_ALPHA, BG_COLOR, BG_ALPHA) -> tuple:
    if FG_COLOR is None:
        fg = None
    else:
        fg = QColor(FG_COLOR)
        fg.setAlpha(FG_ALPHA)
    if BG_COLOR is None:
        bg = None
    else:
        bg = QColor(BG_COLOR)
        bg.setAlpha(BG_ALPHA)

    if fg is None and bg is None:
        return None
    return (fg, bg)

class MonoaRuntime:
    BG_ALPHA = 15
    BG_COLOR = "#FFFF00"
    FG_ALPHA = 200
    FG_COLOR = None
    DEFAULTS = {
        "type": None,
        "range": None,
        "enum": None,
        "color": _default_color(FG_COLOR, FG_ALPHA, BG_COLOR, BG_ALPHA),
        "tooltip": None,
        "label": None,
        "group": None,
        "header": None,
        "title": None,
        "readonly": False,
        "separator": None,
        "space": None
    }
    INSPECTOR_KEYS = ('group', 'header', 'title', 'separator', 'space')
    def __init__(self):
        self.monoa_env = {}
        self.monoa_env['Type'] = self.Type
        self.monoa_env['Range'] = self.Range
        self.monoa_env['Enum'] = self.Enum
        self.monoa_env['Color'] = self.Color
        self.monoa_env['Tooltip'] = self.Tooltip
        self.monoa_env['Label'] = self.Label
        self.monoa_env['Group'] = self.Group
        self.monoa_env['Header'] = self.Header
        self.monoa_env['Title'] = self.Title
        self.monoa_env['Readonly'] = self.Readonly
        self.monoa_env['Separate'] = self.Separator
        self.monoa_env['Separator'] = self.Separator
        self.monoa_env['Space'] = self.Space
        self.monoa_env['Spacer'] = self.Space
        self.cmonoa = None

        # expand
        self.monoa_env.update(self._lower_keyword_expand(self.monoa_env))
        self.monoa_env.update(self._upper_keyword_expand(self.monoa_env))


    @staticmethod
    def _lower_keyword_expand(_d) -> dict:
        return {k.lower(): v for k, v in _d.items()}

    @staticmethod
    def _upper_keyword_expand(_d) -> dict:
        return {k.upper(): v for k, v in _d.items()}

    @staticmethod
    def _remove_builtins(_d) -> dict:
        return {k: v for k, v in _d.items() if not k.startswith('__')}

    def Type(self, x):
        if isinstance(x, str):
            x = x.lower()
            if x == 'int':
                return {"type": int}
            elif x == 'float':
                return {"type": float}
            elif x == 'str':
                return {"type": str}
            elif x == 'bool':
                return {"type": bool}
            else:
                raise FTMonoaRuntimeEvalError(self.cmonoa, f"Type '{x}' is not supported.")
        elif x in (int, float, str, bool):
            return {"type": x}
        raise FTMonoaRuntimeEvalError(self.cmonoa, f"Type '{x}' is not supported.")

    def Range(self, x, y=None, s=1):
        if y is None:
            _range = rangef(0, x, s)
        else:
            # y must be greater than x
            if y <= x:
                raise FTMonoaRuntimeEvalError(self.cmonoa, f"End of range must be greater than start. but got range({x}, {y})")
            _range = rangef(x, y, s)

        # s must be greater than 0
        if s <= 0:
            raise FTMonoaRuntimeEvalError(self.cmonoa, f"Step of range must be greater than 0. but got step({s})")
        return {"range": _range}

    def Enum(self, *args):
        if not args:
            raise FTMonoaRuntimeEvalError(self.cmonoa, f"Enum must have at least one argument.")
        return {"enum": list(args)}

    def Color(self, fg=None, bg=None):
        fg = fg or self.FG_COLOR
        if fg is not None:
            fg = fg.strip()
            if not fg.startswith('#'):
                raise FTMonoaRuntimeUnexpectedColor(self.cmonoa, f"Foreground color must be in #RRGGBB format. but got fg={fg}")
            fg = QColor(fg)
            fg.setAlpha(self.FG_ALPHA)
        bg = bg or self.BG_COLOR
        if bg is not None:
            bg = bg.strip()
            if not bg.startswith('#'):
                raise FTMonoaRuntimeUnexpectedColor(self.cmonoa, f"Background color must be in #RRGGBB format. but got bg={bg}")
            bg = QColor(bg)
            bg.setAlpha(self.BG_ALPHA)
        return {"color": (fg, bg)}

    def Tooltip(self, x):
        return {"tooltip": x}

    def Label(self, x):
        return {"label": x}

    def Group(self, x="", fg=None):
        if x is not None:
            if not isinstance(x, str):
                raise FTMonoaRuntimeEvalError(self.cmonoa, f"Group name must be str. but got group={x}, type={type(x)}")

        fg = fg or self.FG_COLOR
        if fg is not None:
            fg = fg.strip()
            if not fg.startswith('#'):
                raise FTMonoaRuntimeUnexpectedColor(self.cmonoa, f"Foreground color must be in #RRGGBB format. but got fg={fg}")
            fg = QColor(fg)
            fg.setAlpha(self.FG_ALPHA)

        return {"group": (x, fg)}

    def Header(self, x, fg=None):
        if not isinstance(x, str):
            raise FTMonoaRuntimeEvalError(self.cmonoa, f"Header name must be str. but got header={x}, type={type(x)}")

        fg = fg or self.FG_COLOR
        if fg is not None:
            fg = fg.strip()
            if not fg.startswith('#'):
                raise FTMonoaRuntimeUnexpectedColor(self.cmonoa, f"Foreground color must be in #RRGGBB format. but got fg={fg}")
            fg = QColor(fg)
            fg.setAlpha(self.FG_ALPHA)
        return {"header": (x, fg)}

    def Title(self, x, fg=None):
        _ = self.Header(x, fg)
        v = _['header']
        return {"title": v}


    def Readonly(self):
        return {"readonly": True}

    def Separator(self):
        return {"separator": True}

    def Space(self, x=20):
        if x is not None:
            if not isinstance(x, int):
                raise FTMonoaRuntimeEvalError(self.cmonoa, f"Space height must be int. but got space={x}, type={type(x)}")
        return {"space": x}

    def _check_range_with_type(self, _range, _type):
        """
        * 当描述符与Type(int|float)结合并且S的类型不匹配时，会抛出异常
        * 当描述符与Type(str|bool)结合时，会抛出异常
        """
        if _range is None:
            return True
        start, stop, step = _range.start, _range.stop, _range.step
        start, stop, step = start if start is not None else 0, stop, step if step is not None else 1
        if _type == int:
            if not isinstance(start, int):
                raise FTMonoaRuntimeDismatchedRangeType(self.cmonoa, f"Type of range.start must be int. but got range.start={start}")
            if not isinstance(stop, int):
                raise FTMonoaRuntimeDismatchedRangeType(self.cmonoa, f"Type of range.stop must be int. but got range.stop={stop}")
            if not isinstance(step, int):
                raise FTMonoaRuntimeDismatchedRangeType(self.cmonoa, f"Type of range.step must be int. but got range.step={step}")
        elif _type == float:
            if isinstance(start, int):
                start = float(start)
            elif not isinstance(start, float):
                raise FTMonoaRuntimeDismatchedRangeType(self.cmonoa, f"Type of range.start must be float. but got range.start={start}")
            if isinstance(stop, int):
                stop = float(stop)
            elif not isinstance(stop, float):
                raise FTMonoaRuntimeDismatchedRangeType(self.cmonoa, f"Type of range.stop must be float. but got range.stop={stop}")
            if isinstance(step, int):
                step = float(step)
            elif not isinstance(step, float):
                raise FTMonoaRuntimeDismatchedRangeType(self.cmonoa, f"Type of range.step must be float. but got range.step={step}")
        else:
            raise FTMonoaRuntimeEvalError(self.cmonoa, f"Type '{_type}' is not supported for Range.")
        return True


    def _check_enum_with_type(self, _enum, _type, _value):
        """
        * 当描述符与Type(int|float|str)结合时，提升QXXX为QComboBox
        * 当描述符与Type(bool)结合时，会抛出异常
        """
        if _enum is None:
            return True
        if _type == bool:
            raise FTMonoaRuntimeEvalError(self.cmonoa, f"Type '{_type}' is not supported for Enum.")
        elif _type == str:
            if _value not in _enum:
                _enum.insert(0, _value)

        # check each one
        for e in _enum:
            if not isinstance(e, _type):
                raise FTMonoaRuntimeDismatchedEnumType(self.cmonoa, f"Type of Enum must be {_type}. but got Enum={_enum}")

        return True

    def __call__(self, monoa:MonoAttr, monoe:dict) -> tuple[dict, list]:
        """
        解析一个MonoAttr对象，返回一个dict对象，包含了关键单元属性信息; 以及一个list对象，包含了所有的检视信息。
        list对象的每一个元素都是一个tuple对象，包含了检视信息的key和检视信息的内容。
        """
        monoa.iterall(_api_prehandle_single_attr)
        lst, res, self.cmonoa = [], {}, monoa
        env = monoe.copy()
        for expression in monoa:
            try:
                _ = eval(expression, self.monoa_env, env)
            except Exception as e:
                envl = self._remove_builtins(env)
                envg = self._remove_builtins(self.monoa_env)
                raise FTMonoaRuntimeEvalError(self.cmonoa, f"Error while evaluating expression: {expression} - {e}\n\n\tcurrent globals: {str_crop(envg)}\n\tcurrent locals: {str_crop(envl)}")

            # save into res
            if not isinstance(_, dict):
                raise FTMonoaRuntimeEvalError(self.cmonoa, f"Expression '{expression}' must return a dict object.")
            k = list(_)[0]
            if k in self.INSPECTOR_KEYS:
                lst.append((k, _[k]))
            else:
                res.update(_)


        # add name and value
        res['name'] = monoa.name
        res['value'] = monoa.value

        # fill default values
        for k, v in self.DEFAULTS.items():
            if k not in res:
                res[k] = v

        # check type
        if res['type'] is None and monoa.value is None:
            res['type'] = str
            monoa.value = ""
        else:
            is_monoa_value_valid = isinstance(monoa.value, (int, float, str, bool))
            if not is_monoa_value_valid:
                raise FTMonoaRuntimeEvalError(self.cmonoa, f"Type '{type(monoa.value)}' of value '{monoa.value}' is not supported.")
            if res['type'] is None:
                res['type'] = type(monoa.value)
            if res['type'] != type(monoa.value):
                try:
                    monoa.value = res['type'](monoa.value)
                except Exception as e:
                    raise FTMonoaRuntimeEvalError(self.cmonoa, f"Type '{res['type']}' is not supported.")
        # cross check
        self._check_range_with_type(res['range'], res['type'])
        self._check_enum_with_type(res['enum'], res['type'], monoa.value)

        return res, lst


def _api_prehandle_single_attr(attr:str):
    """
    每一条attr都应当为一条python函数执行语句，所以第一步是补全()
    """
    attr = attr.strip()
    # re check
    if attr.endswith(')'):
        return attr
    return attr + '()'


def _api_diagnose_attr_type(obj, monoa:MonoAttr):
    param, value = monoa.name, monoa.value

    for attr in monoa:
        attr = attr.lower()


MONO_FONT = QFont('Consolas', 12, QFont.Bold)
MONO_HEADER_FONT = QFont('Consolas', 13, QFont.Bold)
MONO_TITLE_FONT = QFont('Consolas', 14, QFont.Bold)
MONO_INSPECTOR_FONT = QFont('Consolas', 16, QFont.Bold)

class QMonoWithoutBorder:
    ...

class QMonoRectBorder(QMonoWithoutBorder):
    ...

class QMonoRoundRectBorder(QMonoWithoutBorder):
    ...


class QMonoWidget(QWidget):
    paramChanged = pyqtSignal(str, object)
    def __init__(self, attr_dict:dict, parent=None, *, border=QMonoWithoutBorder):
        super().__init__(parent)
        self.ad = attr_dict  # attr_dict
        self._name = self.ad['name']
        self._value = self.ad['value']
        self._last_emit_value = self._value
        self._rootL = QVBoxLayout()
        self._rootL.setContentsMargins(4, 4, 6, 6)
        self._rootL.setSpacing(4)
        self._mainL = QHBoxLayout()
        self._mainL.setContentsMargins(2, 2, 4, 4)
        self._mainL.setSpacing(2)
        self.setLayout(self._rootL)
        self._rootL.addLayout(self._mainL)
        self._border = border
        assert issubclass(self._border, QMonoWithoutBorder), f"Border must be subclass of QMonoWithoutBorder."

        # ---
        self._uis = []
        self._int_float_vc_flag = False
        self._bool_vc_flag = False
        self._str_vc_flag = False

        self._create_ui()

        # --- set default value
        self._set_default_value()


    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._set_default_value(value=v)

    @property
    def readonly(self):
        return self.ad['readonly']

    def _set_default_value(self, *_, value=None):
        value = value or self.ad['value']
        if self.ad['type'] in (float, int):
            self._int_float_value_changed(value)
        elif self.ad['type'] == bool:
            self._bool_value_changed(value)
        elif self.ad['type'] == str:
            self._str_value_changed(value)
        else:
            raise FTQMonoWidgetRecvObjectError(f"Type of object is not supported in single QMonoWidget.")

    def _select_main_widget(self):
        if self.ad['type'] == int:
            return QSpinBox
        elif self.ad['type'] == float:
            return QDoubleSpinBox
        elif self.ad['type'] == str:
            return QLineEdit
        elif self.ad['type'] == bool:
            return QCheckBox
        elif self.ad['type'] == object:
            raise FTQMonoWidgetRecvObjectError(f"Type of object is not supported in single QMonoWidget.")
        else:
            raise TypeError(f"Type '{self.ad['type']}' is not supported.")

    def _create_ui(self):
        self._lbl = QLabel((self.ad['label'] if self.ad['label'] else self.ad['name']))
        self._lbl.setFont(MONO_FONT)
        self._mainL.addWidget(self._lbl)

        self._mwd = self._select_main_widget()()
        self._mwd.setFont(MONO_FONT)
        self._mainL.addWidget(self._mwd)

        self._btn = QPushButton("O")
        self._btn.setFont(QFont('Arial', 10))
        self._btn.setFixedSize(22, 22)
        self._btn.clicked.connect(self._set_default_value)
        self._mainL.addWidget(self._btn)

        # check range
        if self.ad['range'] and not self.ad['enum']:  # python range
            assert isinstance(self._mwd, (QSpinBox, QDoubleSpinBox)), f"Range is not supported for type '{self.ad['type']}'"
            start, stop, step = self.ad['range'].start, self.ad['range'].stop, self.ad['range'].step
            start, stop, step = start if start is not None else 0, stop, step if step is not None else 1
            self._mwd.setRange(start, stop)
            self._mwd.setSingleStep(step)

            self._qsl = QIntSlider(Qt.Horizontal, start, stop, step) if self.ad['type'] == int else QDoubleSlider(Qt.Horizontal, start, stop, step)
            self._qsl.monoChanged.connect(self._int_float_value_changed)
            self._uis.append(self._qsl)
            self._rootL.addWidget(self._qsl)
        else:
            self._qsl = None

        # check enum
        if self.ad['enum']:
            assert isinstance(self._mwd, (QSpinBox, QDoubleSpinBox, QLineEdit)), f"Enum is not supported for type '{self.ad['type']}'"
            self._mcb = QComboBox()
            self._mcb.setFont(MONO_FONT)
            self._mcb.addItems([str(it) for it in self.ad['enum']])
            self._mcb.currentIndexChanged.connect(self._combo_value_changed)
            self._uis.append(self._mcb)
            self._rootL.addWidget(self._mcb)

            # as readonly
            self._mwd.setReadOnly(True)
        else:
            self._mcb = None

        # posthandle
        self._uis.append(self._lbl)
        self._uis.append(self._mwd)
        self._uis.append(self._btn)

        # transparent
        for ui in self._uis:
            ui.setAttribute(Qt.WA_TranslucentBackground)

        # re label
        if self.ad['label']:
            self._lbl.setText(self.ad['label'])

        # readonly
        if self.ad['readonly']:
            for ui in self._uis:
                # if hasattr(ui, 'setReadOnly'):
                #     ui.setReadOnly(True)
                # else:
                ui.setEnabled(False)

        # tooltips
        if self.ad['tooltip']:
            for ui in self._uis:
                ui.setToolTip(self.ad['tooltip'])

        # color
        if self.ad['color']:  # fg, bg
            fg, bg = self.ad['color']
            readonly = self.ad['readonly']
            if fg is not None:
                fg = QColor(fg.red(), fg.green(), fg.blue(), fg.alpha() if not readonly else int(fg.alpha() * 0.7))
            for ui in self._uis:
                txt = ""
                if fg is not None:
                    txt += f"color:rgba({fg.red()}, {fg.green()}, {fg.blue()}, {fg.alpha()});"
                if not isinstance(ui, (QLineEdit, QSpinBox, QDoubleSpinBox, QPushButton, QComboBox)):
                    # Bg can only be set to other widgets
                    if bg is not None:
                        txt += f"{ui.__class__.__name__}" + "{" + f"background-color:rgba({bg.red()}, {bg.green()}, {bg.blue()}, {bg.alpha()}); + " + "}"
                        txt += f"{ui.__class__.__name__}:focus" + "{" + f"background-color:rgba({bg.red()}, {bg.green()}, {bg.blue()}, {bg.alpha()}); + " + "}"
                if txt:
                    ui.setStyleSheet(txt)

        if isinstance(self._mwd, (QSpinBox, QDoubleSpinBox)):
            self._mwd.valueChanged.connect(self._int_float_value_changed)
        elif isinstance(self._mwd, QCheckBox):
            self._mwd.setText("")
            self._mwd.setStyleSheet("QCheckBox::indicator { width: 20px; height: 20px; };" + self._mwd.styleSheet())
            self._mwd.setMaximumSize(20, 20)
            self._mwd.stateChanged.connect(self._bool_value_changed)
        elif isinstance(self._mwd, QLineEdit):
            self._mwd.textChanged.connect(self._str_value_changed)
    def _int_float_value_changed(self, value, *args):
        if self._int_float_vc_flag: return
        self._int_float_vc_flag = True

        if self._mcb:  # check this first
            if value in self.ad['enum']:
                self._mcb.setCurrentIndex(self.ad['enum'].index(value))
            else:
                idx, value = find_closest(value, self.ad['enum'])
                self._mcb.setCurrentIndex(idx)

        self._mwd.setValue(value)
        if self._qsl:
            self._qsl.setValue(value)

        self._value = value
        self._int_float_vc_flag = False
        self._param_value_changed()

    def _bool_value_changed(self, value):
        if self._bool_vc_flag: return
        self._bool_vc_flag = True

        self._mwd.setChecked(value)
        self._value = bool(value)

        self._bool_vc_flag = False
        self._param_value_changed()

    def _str_value_changed(self, value):
        if self._str_vc_flag: return
        self._str_vc_flag = True

        self._mwd.setText(value)
        self._value = value

        self._str_vc_flag = False
        self._param_value_changed()

    def _param_value_changed(self):
        if self._last_emit_value == self._value:
            return
        self._last_emit_value = self._value
        self.paramChanged.emit(self._name, self._value)


    def _combo_value_changed(self, index):
        value = self.ad['enum'][index]
        if isinstance(self._mwd, (QSpinBox, QDoubleSpinBox)):
            self._int_float_value_changed(value)
        elif isinstance(self._mwd, QLineEdit):
            self._str_value_changed(value)
        else:
            raise TypeError(f"Type '{self.ad['type']}' is not supported with checkbox.")

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        # 获取窗口的大小
        rect = self.rect()
        rect.adjust(0, 0, -4, -4)

        # # 绘制底色
        # base_color = QColor("#888888")
        # base_color.setAlpha(200)
        # painter.setPen(QPen(base_color, 1, Qt.SolidLine))
        # painter.setBrush(Qt.NoBrush)
        # painter.translate(1, 1)  # 偏移
        # painter.drawRoundedRect(rect, 4, 4)
        # base_color.setAlpha(100)
        # painter.translate(1, 1)  # 偏移
        # painter.drawRoundedRect(rect, 4, 4)
        # painter.translate(-2, -2)  # 恢复原位置


        # 设置阴影的颜色和透明度
        shadow_color = QColor("#848080")
        shadow_color.setAlpha(50)
        shadow_fill_color = QColor("#FFFF00") if (self.ad['color'] is None or self.ad['color'][1] is None) else self.ad['color'][1]
        if self.readonly:
            r,g,b = shadow_fill_color.red(), shadow_fill_color.green(), shadow_fill_color.blue()
            average = (r + g + b) // 3
            shadow_fill_color = QColor(int(r * 0.5 + average * 0.5), int(g * 0.5 + average * 0.5), int(b * 0.5 + average * 0.5), shadow_fill_color.alpha())
        shadow_fill_color.setAlpha(25)
        # 融合颜色


        # 绘制底色
        painter.translate(3, 3)
        painter.setPen(QPen(shadow_color, 1, Qt.SolidLine))
        painter.setBrush(QBrush(shadow_fill_color))
        painter.drawRoundedRect(rect, 4, 4)
        painter.translate(-3, -3)  # 恢复原位置

        # 绘制圆角矩形边框
        border_color = QColor("#000000")
        border_color.setAlpha(150)
        painter.setPen(QPen(border_color, 1, Qt.SolidLine))
        painter.setBrush(Qt.NoBrush)
        painter.drawRoundedRect(rect, 4, 4)
        painter.translate(1, 1)
        border_color.setAlpha(100)
        painter.setPen(QPen(border_color, 1, Qt.SolidLine))
        painter.drawRoundedRect(rect, 4, 4)
        painter.translate(1, 1)
        border_color.setAlpha(50)
        painter.setPen(QPen(border_color, 1, Qt.SolidLine))
        painter.drawRoundedRect(rect, 4, 4)
        painter.translate(-2, -2)


        # 在这里添加其他绘制逻辑
        # ...

        # 调用基类的 paintEvent 以进行正常的绘制
        super().paintEvent(event)


class QMonoSeparator(QFrame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)
        self.setLineWidth(1)

class QMonoSpacer(QWidget):
    def __init__(self, height:int=20, parent=None):
        super().__init__(parent)
        self.setFixedHeight(height)

GROUP_INDENT = 32
INSPECTOR_SPACE = 2

class QMonoGroup(QWidget):
    def __init__(self, title: str, color:QColor, parent=None):
        super().__init__(parent)
        icon_color = QColor(color.red() // 4, color.green() // 4, color.blue() // 4, color.alpha())
        self.RIGHT_TRIANGLE = api_create_triangle_pixel(16, icon_color, direct="right")
        self.DOWN_TRIANGLE = api_create_triangle_pixel(16, icon_color, direct="down")
        self._color = color
        if not title.endswith(":"):
            title += ":"
        self._title = title
        self._title_hide = title[:-1]
        self._rootL = QVBoxLayout()
        self._rootL.setContentsMargins(0, 0, 0, 0)
        self._rootL.setSpacing(0)
        self._is_visible = True  # 用于跟踪组件是否可见
        self._title_layout = self._build_title()
        self._setup_header()

        self._mainL = QVBoxLayout()
        self._mainL.setContentsMargins(GROUP_INDENT, 0, 0, 0)
        self._mainL.setSpacing(INSPECTOR_SPACE)

        self._rootL.addLayout(self._mainL)
        self._done_first_paint_flag = False

        self.setLayout(self._rootL)


    @property
    def title(self):
        return self._title

    def _build_title(self):
        _l = QHBoxLayout()
        w = QLabel(self._title)
        w.setFont(MONO_HEADER_FONT)
        w.setStyleSheet(f"color:rgba({self._color.red()}, {self._color.green()}, {self._color.blue()}, {self._color.alpha()});")
        self._label_widget = w
        _l.addWidget(w)
        # add a space
        _l.addStretch()
        return _l

    def _setup_header(self):
        self.toggle_button = QPushButton()
        self.toggle_button.setFixedSize(20, 20)  # 设置按钮大小
        self.toggle_button.setStyleSheet("border: none;")  # 移除按钮边框
        self.toggle_button.clicked.connect(self.toggle_visibility)  # 绑定点击事件

        # 设置按钮的初始方向
        self.toggle_button.setIcon(QIcon(self.DOWN_TRIANGLE))
        self.toggle_button.setIconSize(self.toggle_button.size())

        self.title_layout = QHBoxLayout()
        self.title_layout.addWidget(self.toggle_button)
        self.title_layout.addLayout(self._title_layout)
        self.title_layout.addStretch()

        # 标题布局
        self._rootL.addLayout(self.title_layout)

    def toggle_visibility(self):
        self.setSubVisible(not self._is_visible)

    def setSubVisible(self, visible):
        self._is_visible = visible
        if visible:
            self.toggle_button.setIcon(QIcon(self.DOWN_TRIANGLE))
            # self._rootL.addLayout(self._mainL)
            api_show_layout_widgets(self._mainL)
            self._label_widget.setText(self._title)
        else:
            self.toggle_button.setIcon(QIcon(self.RIGHT_TRIANGLE))
            # self._rootL.removeItem(self._mainL)
            api_hide_layout_widgets(self._mainL)
            self._label_widget.setText(self._title_hide)

    @property
    def main_layout(self):
        return self._mainL

    def firstStart(self):
        self.setSubVisible(False)

    def paintEvent(self, a0):
        if not self._done_first_paint_flag:
            self.firstStart()
            self._done_first_paint_flag = True
        super().paintEvent(a0)
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        if self._is_visible:
            # draw indent line, start at (GROUP_INDENT - 2, self._label_widget.height() + 2)
            _color = QColor(self._color.red(), self._color.green(), self._color.blue(), self._color.alpha() // 2)
            painter.setPen(QPen(_color, 1, Qt.DashLine))
            painter.drawLine(10, self._label_widget.height() + 2, 10, self.height() - 2)
        else:
            # draw after label
            shd_width, pdelta = 48, 12
            geo = self._label_widget.geometry()
            _color = QColor(self._color.red(), self._color.green(), self._color.blue(), self._color.alpha() // 8)
            painter.setPen(Qt.NoPen)
            painter.setBrush(_color)
            shd_rect = QRect(geo.right() + 8, geo.top() + 2, shd_width, geo.height() - 4)
            painter.drawRoundedRect(shd_rect, 2, 2)
            # draw point x3
            _color.setAlpha(_color.alpha() * 4)
            painter.setBrush(_color)
            for i in range(3):
                painter.drawEllipse(QPoint(shd_rect.left() + pdelta * (i + 1), shd_rect.center().y() + 2), 3, 3)



QMONO_INSPECTOR_TITLE_COLOR = QColor(32, 144, 245, 225)

class _QMonoInspector_Area(QWidget):
    rebuildTriggered = pyqtSignal(Mono)
    def __init__(self, mono_target:Mono, parent=None):
        super().__init__(parent)
        self._raw = mono_target.monos
        self.mra = MonoaRuntime()
        self._monos = []
        self._ispts = []
        for m in self._raw:
            mono, ispt = self.mra(m, mono_target.env)
            self._monos.append(mono)
            self._ispts.append(ispt)

        self._mono_widgets = []
        self._group_widgets = {}    # 这是一个按照path来存储group的dict对象 # list[QMonoGroup, dict]
        self._current_group:QWidget = None  # 当前的group
        self._rootL = QVBoxLayout()
        self._rootL.setContentsMargins(2, 2, 0, 2)
        self._rootL.setSpacing(INSPECTOR_SPACE)
        self.setLayout(self._rootL)
        self._create_ui()

        self.rebuildTriggered.connect(self.rebuild)
        self._on_rebuild_flag = False

    @property
    def rebuild_flag(self):
        return self._on_rebuild_flag

    def rebuild(self, mono_target:Mono):
        self._on_rebuild_flag = True
        self._raw = mono_target.monos
        self._monos = []
        self._ispts = []
        for m in self._raw:
            mono, ispt = self.mra(m, mono_target.env)
            self._monos.append(mono)
            self._ispts.append(ispt)

        params_dict = {}
        for mono in self._monos:
            params_dict[mono['name']] = mono['value']

        for qmono in self._mono_widgets:
            if not qmono.readonly:
                continue
            v = params_dict.get(qmono.name, qmono.value)
            qmono.value = v
        self._on_rebuild_flag = False


    @staticmethod
    def _parse_group_path(group_path:str) -> list:
        pat = "[/\\.]"
        return [it.strip() for it in re.split(pat, group_path) if it.strip()]

    def _locate_group(self, group_path:str, color:QColor=None) -> QMonoGroup:
        """
        根据group_path来定位group对象，如果group_path为空，则返回None
        """
        plst = self._parse_group_path(group_path)
        if not plst:
            self._current_group = None
            return None

        # path不为空，因此current必然不为self._group_widgets
        current = self._group_widgets
        self._current_group = None
        for p in plst:
            if p not in current:
                self._set_group(p, color)
                current[p] = [self._current_group, {}]

            self._current_group = current[p][0]
            current = current[p][1]

        return self._current_group

    @property
    def monos(self):
        return self._monos

    def _add_header(self, title:str="", color:QColor=None, *, font=MONO_HEADER_FONT, _sys_widget_type=QLabel, _sys_widget_args=()):
        _l = QHBoxLayout()
        # add a space
        w = QMonoSeparator(self)
        _l.addWidget(w)

        # add a title
        if not title.endswith(":"):
            title += ":"
        w = _sys_widget_type(title, *_sys_widget_args)
        w.setFont(font)
        if color is not None and isinstance(w, QLabel):
            w.setStyleSheet(f"color:rgba({color.red()}, {color.green()}, {color.blue()}, {color.alpha()});")
        _l.addWidget(w)

        # add a space
        w = QMonoSeparator(self)
        _l.addWidget(w)

        # add into
        self._add_layout_into(_l)

    def _add_title(self, title: str = "", color: QColor = None):
        if color is None:
            color = QColor(0, 0, 0, MonoaRuntime.FG_ALPHA)
        return self._add_header(title, color, font=MONO_TITLE_FONT, _sys_widget_type=QShadowLabel, _sys_widget_args=(color, None, 2))


    def _add_separator(self):
        w = QMonoSeparator(self)
        self._add_widget_into(w)

    def _add_space(self, height:int=20):
        w = QMonoSpacer(height)
        self._add_widget_into(w)

    def _set_group(self, title:str, color:QColor=None):
        if color is None:
            color = QColor(0, 0, 0, MonoaRuntime.FG_ALPHA)
        if not title.endswith(":"):
            title += ":"
        w = QMonoGroup(title, color)
        self._add_widget_into(w)
        self._current_group = w

    def _create_inspects(self, isp:list):
        for idx, (k, v) in enumerate(isp):
            if k == 'group' and v is not None:
                if self._current_group is None or self._current_group.title != str(v[0]):
                    self._locate_group(str(v[0]), v[1])
            elif k == 'header' and v is not None:
                self._add_header(*v)  # title, color
            elif k == 'title' and v is not None:
                self._add_title(*v)  # title, color
            elif k == 'space' and v is not None:
                self._add_space(v)
            elif k == 'separator' and v is True:
                self._add_separator()

    def _create_ui(self):
        for i, (m, isp) in enumerate(zip(self._monos, self._ispts)):
            self._create_inspects(isp)
            w = QMonoWidget(m)
            self._mono_widgets.append(w)
            self._add_widget_into(w)

        # add a stretch
        self._rootL.addStretch()

    def _add_widget_into(self, w):
        if self._current_group is None:
            self._rootL.addWidget(w)
        else:
            self._current_group.main_layout.addWidget(w)

    def _add_layout_into(self, l):
        if self._current_group is None:
            self._rootL.addLayout(l)
        else:
            self._current_group.main_layout.addLayout(l)


    @property
    def params(self):
        res = {}
        for w in self._mono_widgets:
            res[w.name] = w.value
        return res

class QMonoInspector(QWidget):
    paramsChanged = pyqtSignal(dict)
    paramChanged = pyqtSignal(str, object)
    # 内部有且只有一个QMonoInspector_Area被一个QScrollArea包裹，只提供一个竖直滚动条
    def __init__(self, mono:Mono, parent=None):
        super().__init__(parent)
        self._mono = mono
        self._inner = _QMonoInspector_Area(mono)
        self._scroll = QScrollArea()
        self._scroll.setWidgetResizable(True)
        self._scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self._scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self._scroll.setFrameShape(QFrame.NoFrame)
        self._scroll.setWidget(self._inner)
        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._add_mono_title("Mono Inspector:")
        self._layout.addWidget(self._scroll)
        self.setLayout(self._layout)

        # connect
        for w in self._inner._mono_widgets:
            w.paramChanged.connect(self._any_param_changed)



    def _any_params_changed(self, name:str, value:object):
        if self._inner.rebuild_flag:
            return
        self.paramsChanged.emit(self.params)

    def _any_param_changed(self, name:str, value:object):
        if self._inner.rebuild_flag:
            return
        self.paramChanged.emit(name, value)
        self._any_params_changed(name, value)


    def _add_mono_title(self, title:str="Mono Inspector:"):
        _l = QHBoxLayout()
        w = QShadowLabel(title, QMONO_INSPECTOR_TITLE_COLOR, None, 2)
        w.setFont(MONO_INSPECTOR_FONT)
        _l.addWidget(w)
        # add a space
        _l.addStretch()
        self._layout.addLayout(_l)


    def rebuildmono(self, s, **params):
        self._mono.handle(s, **params)
        self._inner.rebuild(self._mono.monos)

    @property
    def rebuildTrigger(self):
        return self._inner.rebuildTriggered

    @property
    def monos(self):
        return self._inner.monos

    @property
    def params(self):
        return self._inner.params

    @property
    def qmonos(self):
        return self._inner._mono_widgets.copy()

    def closeEvent(self, a0):
        print("Inspector closed. Params:\n\t", self.params)
        self.close()


def _TestSingleUI(monoa:MonoAttr, monoe:dict):
    mar = MonoaRuntime()
    attr_dict = mar(monoa, monoe)
    print("Parsed :\t")
    for k, v in attr_dict.items():
        print(f"\t{k}:\t{v}")
    app = QApplication(sys.argv)
    w = QMonoWidget(attr_dict)
    w.setWindowTitle("Test")
    w.show()
    sys.exit(app.exec_())

def _TestInspector(mono:Mono):
    app = QApplication(sys.argv)
    w = QMonoInspector(mono)
    w.setWindowTitle("Test")
    w.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    mono = Mono(
        "//\s*>", "\n",
        "\$", "\$",
        "/*\s*>", "\*/",
        "//\s*\[", "\]",
        "//\s*\?", '//\s*:\?', "//\s*:", "//\s*\$",
        COMMENT=r";#", CSO=r"/*\s*#", CEO=r"\*/",
        ENV=r"import math"
    )

    with open('counter.v5.v', 'r', encoding="utf-8") as f:
        test = f.read()

    with open('saved.v', 'w', encoding="utf-8") as f:
        f.write(mono.handle(test, WIDTH=10, add_en=True))

    # test1
    # monos = mono.monos
    # if len(monos) > 0:
    #     one = monos[0]
    #     print(f"mono-attr:\t{one}")
    #     _TestSingleUI(one, mono.env)

    # test2
    _TestInspector(mono)


from PyQt5.QtWidgets import QApplication

from rbpop.win_pop import PopWin
# 控制弹窗的显示


class WinManager:
    def __init__(self):
        self.win_list = []

    @property
    def offset(self):
        # sum of all win's height
        _sum = 0
        for win in self.win_list:
            _sum += win.geometry().height()
        return _sum

    def _start(self):
        for win in self.win_list:
            win.Start()
            win.show()

    def add(self, win:PopWin):
        win.offset = self.offset
        self.win_list.append(win)
        win.all = self.win_list
        win.on_death = lambda: self.remove(win)
        win.Start()
        win.show()

    def remove(self, win:PopWin):
        # try index
        _index = -1
        for i, w in enumerate(self.win_list):
            if w == win:
                _index = i
                break
        if _index == -1:
            return False
        _left, _right = self.win_list[:_index], self.win_list[_index+1:]
        _target = self.win_list[_index]
        _tar_height = _target.geometry().height()
        # win at right's offset should be changed
        for win in _right:
            win.offset -= _tar_height

        # remove
        self.win_list.pop(_index)
        return True

    def hide_all(self):
        for win in self.win_list:
            win.hide()

    def show_all(self):
        for win in self.win_list:
            win.show()


_GLOBAL_WM = [None]

def QPop(pop_win_inst:PopWin):
    if _GLOBAL_WM[0] is None:
        _GLOBAL_WM[0] = WinManager()

    _GLOBAL_WM[0].add(pop_win_inst)



from reft.ft import re, FT, FDict, FSet, FList, FRemove, FTMatched
from reft.mono import Mono
try:
    from reft.mono_ui import QMonoWidget, QMonoInspector
except ImportError:
    QMonoWidget = QMonoInspector = None

__version__ = '0.3.x'





from PyQt5.QtWidgets import QLabel, QApplication
from PyQt5.QtGui import QPainter, QImage, QColor, QPixmap, QPen, QBrush, QFont
from PyQt5.QtCore import Qt, QRect

class QShadowLabel(QLabel):
    def __init__(self, text, text_color:QColor, shadow_color:QColor=None, shadow_radius:int=3, reversed:bool=False, parent=None):
        super(QShadowLabel, self).__init__(parent)
        self._txt_color = text_color
        self._shd_color = shadow_color or QColor(text_color.red() // 4, text_color.green() // 4, text_color.blue() // 4, text_color.alpha())
        self._radius = shadow_radius
        self._reversed = reversed
        self.setText(text)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setPen(Qt.NoPen)
        painter.setBrush(QBrush(Qt.white))

        # 绘制阴影
        _init_alpha = self._shd_color.alpha()
        color = QColor(self._shd_color.red(), self._shd_color.green(), self._shd_color.blue())
        for i in range(self._radius, -1, -1):
            painter.translate(i, i)
            _coef = (i / self._radius) if not self._reversed else (1 - i / self._radius)
            color.setAlpha(int(_init_alpha * _coef + .5))
            painter.setPen(color)
            painter.setFont(self.font())
            painter.drawText(0, 0, self.width(), self.height(), Qt.AlignCenter, self.text())
            painter.translate(-i, -i)

        painter.setPen(QPen(self._txt_color))
        painter.setFont(self.font())
        painter.drawText(0, 0, self.width(), self.height(), Qt.AlignCenter, self.text())
